"""

biogeme.results
===============

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Wed Nov 29 09:44:41 2023
"""
import pandas as pd
from biogeme.version import getText
import biogeme.biogeme as bio
import biogeme.database as db
import biogeme.results as res
from biogeme.expressions import Beta, Variable, exp


# %%
# Version of Biogeme.
print(getText())

# %%
# Definition of a database
df = pd.DataFrame(
    {
        'Person': [1, 1, 1, 2, 2],
        'Exclude': [0, 0, 1, 0, 1],
        'Variable1': [1, 2, 3, 4, 5],
        'Variable2': [10, 20, 30, 40, 50],
        'Choice': [1, 2, 3, 1, 2],
        'Av1': [0, 1, 1, 1, 1],
        'Av2': [1, 1, 1, 1, 1],
        'Av3': [0, 1, 1, 1, 1],
    }
)
df

# %%
my_data = db.Database('test', df)

# %%
# Definition of various expressions
Variable1 = Variable('Variable1')
Variable2 = Variable('Variable2')
beta1 = Beta('beta1', -1.0, -3, 3, 0)
beta2 = Beta('beta2', 2.0, -3, 10, 0)
likelihood = -(beta1**2) * Variable1 - exp(beta2 * beta1) * Variable2 - beta2**4
simul = beta1 / Variable1 + beta2 / Variable2
dict_of_expressions = {'loglike': likelihood, 'beta1': beta1, 'simul': simul}

# %%
# Creation of the BIOGEME object
my_biogeme = bio.BIOGEME(my_data, dict_of_expressions)
my_biogeme.modelName = 'simple_example'
my_biogeme.bootstrap_samples = 10
results = my_biogeme.estimate(run_bootstrap=True)
print(results)

# %%
# Dump results on a file
the_pickle_file = results.writePickle()
print(the_pickle_file)

# %%
# Results can be imported from a file previously generated
read_results = res.bioResults(pickleFile=the_pickle_file)
print(read_results)

# %%
# Results can be formatted in LaTeX
print(read_results.getLaTeX())

# %%
# Results can be formatted in HTML
print(read_results.getHtml())

# %%
# General statistics, including a suggested format.
statistics = read_results.getGeneralStatistics()
statistics

# %%
# The suggested format can be used as follows
for k, (v, p) in statistics.items():
    print(f'{k}:\t{v:{p}}')

# %%
# This result can be generated directly with the following function
print(results.printGeneralStatistics())

# %%
# Estimated parameters as pandas dataframe
read_results.getEstimatedParameters()

# %%
# Correlation results
read_results.getCorrelationResults()

# %%
# Obtain the values of the parameters
read_results.getBetaValues()

# %%
# Obtain the value of one or several specific parameters
read_results.getBetaValues(myBetas=['beta2'])

# %%
# Variance-covariance matrix (Rao-Cramer)
read_results.getVarCovar()

# %%
# Variance-covariance matrix (robust)
read_results.getRobustVarCovar()

# %%
# Variance-covaraince matrix (bootstrap)
read_results.getBootstrapVarCovar()

# %%
# Draws for sensitivity analysis are generated using
# bootstrapping. Any indicator can be generated by the model for each
# draw, and its empirical distribution can be investigate .
read_results.getBetasForSensitivityAnalysis(['beta1', 'beta2'], size=10)

# %%
# Results can be produced in the ALOGIT F12 format
print(read_results.getF12())

# %%
# Miscellaneous functions

# %%
# Likelihood ratio test. Let's first estimate a constrained model

# %%
beta2_constrained = Beta('beta2_constrained', 2.0, -3, 10, 1)
likelihood_constrained = (
    -(beta1**2) * Variable1
    - exp(beta2_constrained * beta1) * Variable2
    - beta2_constrained**4
)
my_biogeme_constrained = bio.BIOGEME(my_data, likelihood_constrained)
my_biogeme_constrained.modelName = 'simple_example_constrained'
results_constrained = my_biogeme_constrained.estimate()
print(results_constrained.short_summary())

# %%
# We can now perform a likelihood ratio test.
test_results = results.likelihood_ratio_test(results_constrained, 0.95)
print(test_results.message)
print(f'Statistic: {test_results.statistic}')
print(f'Threshold: {test_results.threshold}')

# %%
# Calculation of the :math:`p`-value
res.calcPValue(1.96)

# %%
# Compilation of results
dict_of_results = {'Model A': read_results, 'Model B': the_pickle_file}

# %%
df = res.compile_estimation_results(dict_of_results)
df



"""

Example of a catalog
====================

Illustration of the concept of catalog. See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_

:author: Michel Bierlaire, EPFL
:date: Sun Aug  6 18:13:18 2023

"""
import numpy as np
from biogeme import models
from biogeme.expressions import Beta, Variable, Expression
from biogeme.models import boxcox
from biogeme.catalog import (
    Catalog,
    generic_alt_specific_catalogs,
    segmentation_catalogs,
)
from biogeme.nests import OneNestForNestedLogit, NestsForNestedLogit

# %%
# See :ref:`swissmetro_data`
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)


# %%
# Function printing all configurations of an expression.
def print_all_configurations(expression: Expression) -> None:
    """Prints all configurations that an expression can take"""
    expression.set_central_controller()
    total = expression.central_controller.number_of_configurations()
    print(f'Total: {total} configurations')
    for config_id in expression.central_controller.all_configurations_ids:
        print(config_id)


# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob_logit = models.loglogit(V, av, CHOICE)

# %%
# Nest definition.

mu_existing = Beta('mu_existing', 1, 1, 10, 0)
existing = OneNestForNestedLogit(nest_param=mu_existing, list_of_alternatives=[1, 3])
nests = NestsForNestedLogit(choice_set=list(V), tuple_of_nests=(existing,))

# %%
# Contribution to the log-likelihood.
logprob_nested = models.lognested(V, av, nests, CHOICE)

# %%
# Definition of the catalog containnig two models specifications:
# logit and nested logit.
model_catalog = Catalog.from_dict(
    catalog_name='model_catalog',
    dict_of_expressions={
        'logit': logprob_logit,
        'nested': logprob_nested,
    },
)

# %%
# Current status of the catalog.
print(model_catalog)

# %%
# Use the controller to select a different configuration.
model_catalog.controlled_by.set_name('nested')
print(model_catalog)

# %%
# Iterator.
for specification in model_catalog:
    print(specification)

# %%
# All configurations.
print_all_configurations(model_catalog)

# %% Non linear specifications.
TRAIN_TT = Variable('TRAIN_TT')
TRAIN_COST = Variable('TRAIN_COST')
ell_travel_time = Beta('lambda_travel_time', 1, -10, 10, 0)
linear_train_tt = TRAIN_TT
boxcox_train_tt = boxcox(TRAIN_TT, ell_travel_time)
squared_train_tt = TRAIN_TT * TRAIN_TT
train_tt_catalog = Catalog.from_dict(
    catalog_name='train_tt_catalog',
    dict_of_expressions={
        'linear': linear_train_tt,
        'boxcox': boxcox_train_tt,
        'squared': squared_train_tt,
    },
)

# %%
# Define a utility function involving the catalog.
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, 0, 0)
V_TRAIN = ASC_TRAIN + B_TIME * train_tt_catalog

# %%
print_all_configurations(V_TRAIN)

# %%
# Unsynchronized catalogs
CAR_TT = Variable('CAR_TT')
CAR_COST = Variable('CAR_COST')
linear_car_tt = CAR_TT
boxcox_car_tt = boxcox(CAR_TT, ell_travel_time)
squared_car_tt = CAR_TT * CAR_TT
car_tt_catalog = Catalog.from_dict(
    catalog_name='car_tt_catalog',
    dict_of_expressions={
        'linear': linear_car_tt,
        'boxcox': boxcox_car_tt,
        'squared': squared_car_tt,
    },
)

# %%
# Create a dummy expression with the two catalogs.
dummy_expression = train_tt_catalog + car_tt_catalog

# %%
print_all_configurations(dummy_expression)

# %%
# Synchronized catalogs.
CAR_TT = Variable('CAR_TT')
CAR_COST = Variable('CAR_COST')
linear_car_tt = CAR_TT
boxcox_car_tt = boxcox(CAR_TT, ell_travel_time)
squared_car_tt = CAR_TT * CAR_TT
car_tt_catalog = Catalog.from_dict(
    catalog_name='car_tt_catalog',
    dict_of_expressions={
        'linear': linear_car_tt,
        'boxcox': boxcox_car_tt,
        'squared': squared_car_tt,
    },
    controlled_by=train_tt_catalog.controlled_by,
)

# %%
# Create a dummy expression with the two catalogs.
dummy_expression = train_tt_catalog + car_tt_catalog

# %%
print_all_configurations(dummy_expression)

# %%
# Alternative specific specification.

(B_TIME_catalog_dict, B_COST_catalog_dict) = generic_alt_specific_catalogs(
    generic_name='coefficients',
    beta_parameters=[B_TIME, B_COST],
    alternatives=('TRAIN', 'CAR'),
)

# %%
# Create utility functions involving those catalogs.
V_TRAIN = (
    B_TIME_catalog_dict['TRAIN'] * TRAIN_TT + B_COST_catalog_dict['TRAIN'] * TRAIN_COST
)
V_CAR = B_TIME_catalog_dict['CAR'] * CAR_TT + B_COST_catalog_dict['CAR'] * CAR_COST

# %%
# Create a dummy expression involving the utility functions.
dummy_expression = V_TRAIN + V_CAR

# %%
print_all_configurations(dummy_expression)

# %%
# Alternative specific - not synchronized.

(B_TIME_catalog_dict,) = generic_alt_specific_catalogs(
    generic_name='time_coefficient',
    beta_parameters=[B_TIME],
    alternatives=('TRAIN', 'CAR'),
)

(B_COST_catalog_dict,) = generic_alt_specific_catalogs(
    generic_name='cost_coefficient',
    beta_parameters=[B_COST],
    alternatives=('TRAIN', 'CAR'),
)

# %%
# Create utility functions involving those catalogs.
V_TRAIN = (
    B_TIME_catalog_dict['TRAIN'] * TRAIN_TT + B_COST_catalog_dict['TRAIN'] * TRAIN_COST
)
V_CAR = B_TIME_catalog_dict['CAR'] * CAR_TT + B_COST_catalog_dict['CAR'] * CAR_COST

# %%
# Create a dummy expression involving the utility functions.
dummy_expression = V_TRAIN + V_CAR

# %%
print_all_configurations(dummy_expression)

# %%
# Segmentation

# %%
# We consider two trip purposes: `commuters` and anything else. We
# need to define a binary variable first.
database.data['COMMUTERS'] = np.where(database.data['PURPOSE'] == 1, 1, 0)

# %%
# Segmentation on trip purpose.
segmentation_purpose = database.generate_segmentation(
    variable='COMMUTERS',
    mapping={0: 'non_commuters', 1: 'commuters'},
    reference='non_commuters',
)

# %%
# Segmentation on luggage.
segmentation_luggage = database.generate_segmentation(
    variable='LUGGAGE',
    mapping={0: 'no_lugg', 1: 'one_lugg', 3: 'several_lugg'},
    reference='no_lugg',
)

# %%
# Catalog of segmented alternative specific constants, allows a maximum
# of two segmentations.
ASC_TRAIN_catalog, ASC_CAR_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_TRAIN, ASC_CAR],
    potential_segmentations=(
        segmentation_purpose,
        segmentation_luggage,
    ),
    maximum_number=2,
)

# %%
# Create a dummy expression.
dummy_expression = ASC_TRAIN_catalog + ASC_CAR_catalog

# %%
print_all_configurations(dummy_expression)

# %%
# Catalog of segmented alternative specific constants, allows a maximum
# of one segmentation.
ASC_TRAIN_catalog, ASC_CAR_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_TRAIN, ASC_CAR],
    potential_segmentations=(
        segmentation_purpose,
        segmentation_luggage,
    ),
    maximum_number=1,
)

# %%
# Create a dummy expression.
dummy_expression = ASC_TRAIN_catalog + ASC_CAR_catalog

# %%
print_all_configurations(dummy_expression)

# %%
# Segmentation and alternative specific
# Maximum one segmentation.
(B_TIME_catalog_dict,) = generic_alt_specific_catalogs(
    generic_name='B_TIME',
    beta_parameters=[B_TIME],
    alternatives=['TRAIN', 'CAR'],
    potential_segmentations=(
        segmentation_purpose,
        segmentation_luggage,
    ),
    maximum_number=1,
)

# %%
print_all_configurations(B_TIME_catalog_dict['TRAIN'])

# %%
# Maximum two segmentations.
(B_TIME_catalog_dict,) = generic_alt_specific_catalogs(
    generic_name='B_TIME',
    beta_parameters=[B_TIME],
    alternatives=['TRAIN', 'CAR'],
    potential_segmentations=(
        segmentation_purpose,
        segmentation_luggage,
    ),
    maximum_number=2,
)

# %%
print_all_configurations(B_TIME_catalog_dict['TRAIN'])



import unittest

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Numeric, Expression
from biogeme.mdcev import NonMonotonic


class TestNonMonotonicMdcev(unittest.TestCase):

    def setUp(self):
        self.model_name = 'TestModel'
        self.baseline_utilities = {1: Numeric(1), 2: Numeric(2)}
        self.gamma_parameters = {1: Numeric(0.5), 2: None}
        self.alpha_parameters = {1: Numeric(0.1), 2: Numeric(0.5)}
        self.scale_parameter = Numeric(0.1)
        self.weights = Numeric(1.0)
        self.instance = NonMonotonic(
            model_name=self.model_name,
            baseline_utilities=self.baseline_utilities,
            mu_utilities=self.baseline_utilities,
            gamma_parameters=self.gamma_parameters,
            alpha_parameters=self.alpha_parameters,
            scale_parameter=self.scale_parameter,
            weights=self.weights,
        )

    def test_one_alternative_with_gamma(self):
        the_id = 1
        the_consumption = Numeric(3)
        expected_value = 1.471743848695463
        utility = self.instance.transformed_utility(the_id, the_consumption)
        self.assertIsInstance(utility, Expression)
        self.assertAlmostEqual(utility.get_value(), expected_value, places=3)
        log_determinant = self.instance.calculate_log_determinant_one_alternative(
            the_id, the_consumption
        )
        self.assertAlmostEqual(
            log_determinant.get_value(), -2.109442618302976, places=3
        )
        inverse_determinant = (
            self.instance.calculate_inverse_of_determinant_one_alternative(
                the_id, the_consumption
            )
        )
        self.assertAlmostEqual(
            inverse_determinant.get_value(), 8.243645146922482, places=3
        )

    def test_utility_one_alternative_without_gamma(self):
        the_id = 2
        the_consumption = Numeric(3)
        utility = self.instance.transformed_utility(the_id, the_consumption)
        self.assertAlmostEqual(utility.get_value(), 6.266073527774857, places=3)
        log_determinant = self.instance.calculate_log_determinant_one_alternative(
            the_id, the_consumption
        )
        self.assertAlmostEqual(
            log_determinant.get_value(), -0.34106561356210996, places=3
        )
        inverse_determinant = (
            self.instance.calculate_inverse_of_determinant_one_alternative(
                the_id, the_consumption
            )
        )
        self.assertAlmostEqual(
            inverse_determinant.get_value(), 1.4064455197352266, places=3
        )

    def test_calculate_log_determinant_one_alternative_error(self):
        self.assertRaises(
            BiogemeError,
            self.instance.calculate_log_determinant_one_alternative,
            the_id=10,
            consumption=Numeric(2),
        )



"""
This module defines the MultiRowEvaluator class, which evaluates multiple expressions
on a given database using JAX for efficient batched computation. It returns results
as a pandas DataFrame with one column per expression and one row per observation.

Michel Bierlaire
Wed Apr 2 13:10:17 2025
"""

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd

from biogeme.exceptions import BiogemeError
from biogeme.expressions import build_vectorized_function
from biogeme.floating_point import JAX_FLOAT, NUMPY_FLOAT
from biogeme.model_elements import ModelElements


class MultiRowEvaluator:
    """
    Evaluates multiple expressions on a common dataset using JAX and returns results
    as a pandas DataFrame. This class compiles all expressions into JAX functions and
    evaluates them efficiently in a single batched operation.

    :param model_elements: Object containing the expressions and all elements needed to calculate them.
    """

    def __init__(self, model_elements: ModelElements, numerically_safe: bool):
        if model_elements is None:
            raise BiogemeError('A model must be provided.')
        self.multiple_model_elements = model_elements
        self.free_betas_names = model_elements.expressions_registry.free_betas_names
        self.data_jax = model_elements.database.data_jax
        self.draws_jax = model_elements.draws_management.draws_jax
        self.names = list(model_elements.expressions.keys())
        n_rv = (
            self.multiple_model_elements.expressions_registry.number_of_random_variables
        )
        self.random_variables_jax = jnp.zeros((n_rv,), dtype=JAX_FLOAT)

        self.vectorized_functions = [
            build_vectorized_function(
                expr.recursive_construct_jax_function(numerically_safe=numerically_safe)
            )
            for expr in self.multiple_model_elements.expressions.values()
        ]

        @jax.jit
        def evaluate_all(params, data, draws, rv):
            return jnp.stack(
                [vf(params, data, draws, rv) for vf in self.vectorized_functions],
                axis=1,
            )

        self._evaluate_all = evaluate_all

    def evaluate(self, the_betas: dict[str, float]) -> pd.DataFrame:
        """
        Evaluates all expressions using the provided beta values.

        :param the_betas: A dictionary mapping beta names to their numerical values.
        :return: A pandas DataFrame with one column per expression and one row per observation.
        """
        param_vector = (
            self.multiple_model_elements.expressions_registry.get_betas_array(the_betas)
        )
        values = self._evaluate_all(
            param_vector, self.data_jax, self.draws_jax, self.random_variables_jax
        )
        return pd.DataFrame(np.asarray(values, dtype=NUMPY_FLOAT), columns=self.names)



"""Defines  a catalog of expressions that may be considered in a specification

Michel Bierlaire
Wed Apr 16 18:35:02 2025

"""

from __future__ import annotations

import logging
from typing import Iterator

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Expression,
    MultipleExpression,
    NamedExpression,
    validate_and_convert,
)

from .controller import Controller

logger = logging.getLogger(__name__)


class Catalog(MultipleExpression):
    """Catalog of expressions that are interchangeable. Only one of
    them defines the specification. They are designed to be
    modified algorithmically by a controller.
    """

    def __init__(
        self,
        catalog_name: str,
        named_expressions: list[NamedExpression],
        controlled_by: Controller | None = None,
    ):
        """Ctor

        :param catalog_name: name of the catalog of expressions

        :param named_expressions: list of NamedExpression,
            each containing a name and an expression.

        :param controlled_by: Object controlling the selection of the specifications.

        :raise BiogemeError: if list_of_named_expressions is empty
        :raise BiogemeError: if incompatible Controller

        """
        super().__init__(catalog_name)

        if not named_expressions:
            raise BiogemeError(
                f'{catalog_name}: cannot create a catalog from an empty list.'
            )

        if controlled_by and not isinstance(controlled_by, Controller):
            error_msg = (
                f'The controller must be of type Controller and not '
                f'{type(controlled_by)}'
            )
            raise BiogemeError(error_msg)

        self.named_expressions = [
            NamedExpression(
                name=named.name,
                expression=validate_and_convert(named.expression),
            )
            for named in named_expressions
        ]

        # Declare the expressions as children of the catalog
        for named_expression in self.named_expressions:
            self.children.append(named_expression.expression)

        names = [named_expr.name for named_expr in self.named_expressions]
        if controlled_by is None:
            controller_name = catalog_name
            self.controlled_by = Controller(
                controller_name=controller_name, specification_names=names
            )
        else:
            self.controlled_by = controlled_by
            controller_names = list(controlled_by.specification_names)
            if names != controller_names:
                error_msg = (
                    f'Incompatible IDs between catalog [{names}] and controller '
                    f'[{controller_names}]'
                )
                raise BiogemeError(error_msg)

    def get_all_controllers(self) -> set[Controller]:
        """Provides all controllers controlling the specifications of
            a multiple expression

        :return: a set of controllers
        :rtype: set(biogeme.controller.Controller)

        """
        all_controllers = {self.controlled_by}
        for e in self.children:
            all_controllers |= e.get_all_controllers()
        return all_controllers

    @classmethod
    def from_dict(
        cls,
        catalog_name: str,
        dict_of_expressions: dict[str, Expression],
        controlled_by: Controller | None = None,
    ) -> Catalog:
        """Ctor using a dict instead of a list.

        Python used not to guarantee the order of elements of a dict,
        although, in practice, it is always preserved. If the order is
        critical, it is better to use the main constructor. If not,
        this constructor provides a more readable code.

        :param catalog_name: name of the catalog
        :param dict_of_expressions: dict associating the name of an
            expression and the expression itself.
        :param controlled_by: Object controlling the selection of the specifications.

        """
        named_expressions = [
            NamedExpression(name=name, expression=expression)
            for name, expression in dict_of_expressions.items()
        ]
        return cls(
            catalog_name=catalog_name,
            named_expressions=named_expressions,
            controlled_by=controlled_by,
        )

    def __iter__(self) -> Iterator[NamedExpression]:
        """Obtain an iterator on the named expressions"""
        return self.get_iterator()

    def get_iterator(self) -> Iterator[NamedExpression]:
        """Obtain an iterator on the named expressions"""
        return iter(self.named_expressions)

    def selected(self) -> NamedExpression:
        """Return the selected expression and its name

        :return: the name and the selected expression
        :rtype: NamedExpression
        """
        return self.named_expressions[self.controlled_by.current_index]



"""Tests for parameter expression

Michel Bierlaire
Tue Mar 25 18:33:51 2025
"""

import unittest

import jax.numpy as jnp

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, TypeOfElementaryExpression


class TestBetaParameter(unittest.TestCase):
    def test_constructor_valid(self):
        beta = Beta('beta1', 1.0, 0.0, 10.0, 0)
        self.assertEqual(beta.name, 'beta1')
        self.assertEqual(beta.init_value, 1.0)
        self.assertEqual(beta.lower_bound, 0.0)
        self.assertEqual(beta.upper_bound, 10.0)
        self.assertEqual(beta.status, 0)
        self.assertIsNone(beta.specific_id)

    def test_constructor_invalid_name(self):
        with self.assertRaises(BiogemeError):
            Beta(123, 1.0, None, None, 0)

    def test_constructor_invalid_value(self):
        with self.assertRaises(BiogemeError):
            Beta('beta1', 'invalid', None, None, 0)

    def test_expression_type(self):
        beta_free = Beta('beta1', 1.0, None, None, 0)
        self.assertEqual(
            beta_free.expression_type, TypeOfElementaryExpression.FREE_BETA
        )
        beta_fixed = Beta('beta2', 1.0, None, None, 1)
        self.assertEqual(
            beta_fixed.expression_type, TypeOfElementaryExpression.FIXED_BETA
        )

    def test_str_and_repr(self):
        beta = Beta('beta1', 1.0, None, None, 0)
        self.assertIn("Beta('beta1', 1.0", str(beta))
        self.assertIn('<Beta name=beta1 value=1.0', repr(beta))

    def test_safe_beta_id_raises(self):
        beta = Beta('beta1', 1.0, None, None, 0)
        with self.assertRaises(BiogemeError):
            _ = beta.safe_beta_id

    def test_safe_beta_id_returns(self):
        beta = Beta('beta1', 1.0, None, None, 0)
        beta.specific_id = 5
        self.assertEqual(beta.safe_beta_id, 5)

    def test_fix_betas_changes_value_and_status(self):
        beta = Beta('beta1', 1.0, None, None, 0)
        beta.fix_betas({'beta1': 2.5})
        self.assertEqual(beta.init_value, 2.5)
        self.assertEqual(beta.status, 1)

    def test_fix_betas_renames(self):
        beta = Beta('beta1', 1.0, None, None, 0)
        beta.fix_betas({'beta1': 2.5}, prefix='pre_', suffix='_suf')
        self.assertEqual(beta.name, 'pre_beta1_suf')

    def test_change_init_values_free(self):
        beta = Beta('beta1', 1.0, None, None, 0)
        beta.change_init_values({'beta1': 3.0})
        self.assertEqual(beta.init_value, 3.0)

    def test_change_init_values_fixed_logs_warning(self):
        beta = Beta('beta1', 1.0, None, None, 1)
        beta.change_init_values({'beta1': 2.0})
        self.assertEqual(beta.init_value, 2.0)

    def test_jax_function_returns_value(self):
        beta = Beta('beta1', 1.0, None, None, 0)
        beta.specific_id = 2
        fn = beta.recursive_construct_jax_function(numerically_safe=False)
        parameters = jnp.array([0.0, 1.0, 3.14])
        result = fn(parameters, None, None, None)
        self.assertEqual(result, 3.14)



"""

Re-estimation of best models
============================

After running the assisted specification algorithm for the 432
specifications in :ref:`everything_spec_section`, we use post processing to
re-estimate all Pareto optimal models, and display some information
about the algorithm.
See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_.

:author: Michel Bierlaire, EPFL
:date: Thu Jul 20 17:15:37 2023

"""
try:
    import matplotlib.pyplot as plt

    can_plot = True
except ModuleNotFoundError:
    can_plot = False
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme.assisted import ParetoPostProcessing

from everything_spec import model_catalog, database

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b08selected_specification')

PARETO_FILE_NAME = 'saved_results/b07everything_assisted.pareto'

# %%
# Create the biogeme object from the catalog.
the_biogeme = bio.BIOGEME(database, model_catalog)
the_biogeme.modelName = 'b09post_processing'

# %%
# Create the post processing object.
post_processing = ParetoPostProcessing(
    biogeme_object=the_biogeme, pareto_file_name=PARETO_FILE_NAME
)

# %%
# Re-estimate the models.
all_results = post_processing.reestimate(recycle=True)

# %%
# We retieve the first estimation results for illustration.
spec, results = next(iter(all_results.items()))

# %%
print(spec)

# %%
print(results.short_summary())

# %%
results.getEstimatedParameters()


# %%
# The following plot illustrates all models that have been
# estimated.  Each dot corresponds to a model. The x-coordinate
# corresponds to the Akaike Information Criterion (AIC). The
# y-coordinate corresponds to the Bayesian Information Criterion
# (BIC). Note that there is a third objective that does not appear on
# this picture: the number of parameters. If the shape of the dot is a
# circle, it means that it corresponds to a Pareto optimal model. If
# the shape is a cross, it means that the model has been Pareto
# optimal at some point during the algorithm and later removed as a
# new model dominating it has been found. If the shape is a start, it
# means that the model has been deemed invalid.
if can_plot:
    _ = post_processing.plot(
        label_x='Nbr of parameters',
        label_y='Negative log likelihood',
        objective_x=1,
        objective_y=0,
    )



"""

Examples of mathematical expressions
====================================

Example of manipulating mathematical expressions and calculation of
derivatives.

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:06:21 2023

"""
import numpy as np

try:
    import matplotlib.pyplot as plt

    can_plot = True
except ModuleNotFoundError:
    can_plot = False


from biogeme.expressions import Beta, exp

# ##
# We create a simple expression:
b = Beta('b', 1, None, None, 0)
expression = exp(-b * b + 1)

# %%
# We can calculate its value. Note that, as the expression is
# calculated out of Biogeme, the IDs must be prepared. So the
# parameter 'prepareIds' is set to True
z = expression.getValue_c(prepareIds=True)
print(f'exp(-b * b + 1) = {z}')

# %%
# We can also calculate the value, the first derivative, the second
# derivative, and the BHHH, which in this case is the square of the
# first derivatives
f, g, h, bhhh = expression.getValueAndDerivatives(prepareIds=True)
# %%
print(f'f = {f}')
# %%
print(f'g = {g}')
# %%
print(f'h = {h}')
# %%
print(f'BHHH = {bhhh}')

# %%
# From the expression, we can create a Python function that takes as
# argument the value of the free parameters, and returns the function,
# the first and second derivatives.
fct = expression.createFunction()

# %%
# We can use the function for different values of the parameter
beta = 2
f, g, h = fct(beta)
print(f'f({beta}) = {f}')
print(f'g({beta}) = {g}')
print(f'h({beta}) = {h}')

# %%
beta = 3
f, g, h = fct(beta)
print(f'f({beta}) = {f}')
print(f'g({beta}) = {g}')
print(f'h({beta}) = {h}')

# %%
if can_plot:
    # We can also use it to plot the function and its derivatives
    x = np.arange(-2, 2, 0.01)

    # The value of the function is element [0].
    f = [fct(xx)[0] for xx in x]

    # The gradient is element [1]. As it contains only one entry [0],
    # we convert it into float.
    g = [float(fct([xx])[1][0]) for xx in x]

    # The hessian is element [2]. As it contains only one entry
    # [0][0], we convert it into float.
    h = [float(fct([xx])[2][0][0]) for xx in x]

    ax = plt.gca()
    ax.plot(x, f, label="f(x)")
    ax.plot(x, g, label="f'(x)")
    ax.plot(x, h, label='f"(x)')
    ax.legend()

    plt.show()



"""
Test the optimization module

:author: Michel Bierlaire
:data: Wed Apr 29 17:45:19 2020
"""

# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring


import random as rnd
import unittest

import numpy as np

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, Variable
from biogeme.parameters import Parameters
from test_data import getData


class test_optimization(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)
        rnd.seed(90267)
        Choice = Variable('Choice')
        Variable1 = Variable('Variable1')
        Variable2 = Variable('Variable2')
        beta1 = Beta('beta1', 0, None, None, 0)
        beta2 = Beta('beta2', 0, None, None, 0)
        V1 = beta1 * Variable1
        V2 = beta2 * Variable2
        V3 = 0
        V = {1: V1, 2: V2, 3: V3}

        self.likelihood = models.loglogit(V, av=None, i=Choice)

        self.scipy_configuration = Parameters()
        self.scipy_configuration.set_value(
            name='optimization_algorithm', value='scipy', section='Estimation'
        )
        self.ls_configuration = Parameters()
        self.ls_configuration.set_value(
            name='optimization_algorithm', value='LS-newton', section='Estimation'
        )
        self.tr_configuration = Parameters()
        self.tr_configuration.set_value(
            name='optimization_algorithm', value='TR-newton', section='Estimation'
        )
        self.simple_bounds_configuration = Parameters()
        self.simple_bounds_configuration.set_value(
            name='optimization_algorithm', value='simple_bounds', section='Estimation'
        )

    def testBioScipy(self):
        my_biogeme = bio.BIOGEME(
            getData(1),
            self.likelihood,
            parameters=self.scipy_configuration,
            generate_html=False,
            generate_yaml=False,
            save_iterations=False,
        )
        my_biogeme.model_name = 'simpleExample'
        results = my_biogeme.estimate()
        beta = results.get_beta_values()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)

    def _testBioNewtonLineSearch(self):
        my_biogeme = bio.BIOGEME(
            getData(1),
            self.likelihood,
            parameters=self.ls_configuration,
            generate_html=False,
            generate_yaml=False,
            save_iterations=False,
        )
        my_biogeme.model_name = 'simpleExample'
        results = my_biogeme.estimate()
        beta = results.get_beta_values()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)

    def testBioNewtonTrustRegion(self):
        my_biogeme = bio.BIOGEME(
            getData(1),
            self.likelihood,
            parameters=self.tr_configuration,
            generate_html=False,
            generate_yaml=False,
            save_iterations=False,
        )
        my_biogeme.model_name = 'simpleExample'
        results = my_biogeme.estimate()
        beta = results.get_beta_values()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)

    def testBioNewtonSimpleBounds(self):
        my_biogeme = bio.BIOGEME(
            getData(1),
            self.likelihood,
            parameters=self.simple_bounds_configuration,
            generate_html=False,
            generate_yaml=False,
            save_iterations=False,
        )
        my_biogeme.model_name = 'simpleExample'
        results = my_biogeme.estimate()
        beta = results.get_beta_values()
        self.assertAlmostEqual(beta['beta1'], 0.144546, 3)
        self.assertAlmostEqual(beta['beta2'], 0.023502, 3)


if __name__ == '__main__':
    unittest.main()



"""Implementation of the "gamma profile" MDCEV model. See the technical report.

Michel Bierlaire
Sun Apr 7 17:16:53 2024
"""

import logging

import numpy as np

from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, log, exp, Numeric
from biogeme.mdcev.mdcev import Mdcev
from biogeme.tools.checks import validate_dict_types

logger = logging.getLogger(__name__)


class GammaProfile(Mdcev):
    def __init__(
        self,
        model_name: str,
        baseline_utilities: dict[int, Expression],
        gamma_parameters: dict[int, Expression | None],
        alpha_parameters: dict[int, Expression] | None = None,
        scale_parameter: Expression | None = None,
        prices: dict[int, Expression] | None = None,
        weights: Expression | None = None,
    ) -> None:
        """Implementation of the MDCEV model with Gamma profile."""
        super().__init__(
            model_name=model_name,
            baseline_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameters,
            scale_parameter=scale_parameter,
            weights=weights,
        )
        self.prices: dict[int, Expression] | None = prices
        if self.prices is not None:
            validate_dict_types(self.prices, 'prices', Expression)

    def transformed_utility(
        self,
        the_id: int,
        the_consumption: Expression,
    ) -> Expression:
        """Calculates the utility for one alternative. This is the term V needed in the loglikelihood function"""

        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma: Expression | None = self.gamma_parameters[the_id]
        price: Expression | None = None if self.prices is None else self.prices[the_id]

        baseline_utility = self.baseline_utilities[the_id]

        if gamma is None:
            return baseline_utility - log(the_consumption)

        if price is None:
            return baseline_utility + log(gamma) - log(the_consumption + gamma)

        return baseline_utility + log(gamma) - log(the_consumption + price * gamma)

    def calculate_log_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the log of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma: Expression | None = self.gamma_parameters.get(the_id)

        if gamma is None:
            return -log(consumption)
        if self.prices is None:
            return -log(consumption + gamma)
        price = self.prices[the_id]
        return -log(consumption + price * gamma)

    def calculate_inverse_of_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the inverse of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma: Expression | None = self.gamma_parameters[the_id]

        if gamma is None:
            return consumption
        if self.prices is None:
            return consumption + gamma
        price = self.prices[the_id]
        return consumption + price * gamma

    def utility_expression_one_alternative(
        self,
        the_id: int,
        the_consumption: Expression,
        unscaled_epsilon: Expression,
    ) -> Expression:
        """Utility expression. Used only for code validation."""
        baseline_utility = self.baseline_utilities[the_id]
        epsilon: Expression = (
            unscaled_epsilon
            if self.scale_parameter is None
            else unscaled_epsilon / self.scale_parameter
        )
        price: Expression = Numeric(1) if self.prices is None else self.prices[the_id]
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return exp(baseline_utility + epsilon) * log(the_consumption / price)
        return (
            exp(baseline_utility + epsilon)
            * gamma
            * log(1 + the_consumption / (price * gamma))
        )

    def utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Utility used in the optimization problem solved for forecasting."""
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None and np.isclose(the_consumption, 0.0):
            error_msg = f'Alternative {the_id} is the outside good. Its consumption cannot be zero.'
            raise BiogemeError(error_msg)
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        price: float = 1 if self.prices is None else self.prices[the_id].get_value()

        if gamma is None:
            return np.exp(baseline_utility + epsilon) * np.log(the_consumption / price)
        return (
            np.exp(baseline_utility + epsilon)
            * gamma.get_value()
            * np.log(1 + the_consumption / (price * gamma.get_value()))
        )

    def derivative_utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Used in the optimization problem solved for forecasting to calculate the dual variable."""

        # For the outside good, the value at zero  consumption is +infinity.
        if the_id == self.outside_good_index and the_consumption == 0.0:
            return np.inf

        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        price: float = 1 if self.prices is None else self.prices[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return np.exp(baseline_utility + epsilon) / the_consumption
        return (
            np.exp(baseline_utility + epsilon)
            * gamma.get_value()
            / (the_consumption + price * gamma.get_value())
        )

    def optimal_consumption_one_alternative(
        self,
        the_id: int,
        dual_variable: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Analytical calculation of the optimal consumption if the dual variable is known."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        price: float = 1 if self.prices is None else self.prices[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return np.exp(baseline_utility + epsilon) / dual_variable
        return (
            np.exp(baseline_utility + epsilon) * gamma.get_value() / dual_variable
            - price * gamma.get_value()
        )

    def lower_bound_dual_variable(
        self,
        chosen_alternatives: set[int],
        one_observation: Database,
        epsilon: np.ndarray,
    ) -> float:
        """Method providing model specific bounds on the dual variable. It not overloaded,
        default values are used.

        :param chosen_alternatives: list of alternatives that are chosen at the optimal solution
        :param one_observation: data for one observation.
        :param epsilon: draws from the error term.
        :return: a lower bound on the dual variable, such that the expenditure calculated for any larger value is
        well-defined and non negative.
        """
        return 0.0

    def _list_of_expressions(self) -> list[Expression]:
        """Extract the list of expressions involved in the model"""
        if self.prices is None:
            return []
        return [expression for expression in self.prices.values()]



"""Function collecting recursively information about expressions

Michel Bierlaire
Thu May 01 2025, 18:50:24
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from .beta_parameters import Beta
from .draws import Draws
from .random_variable import RandomVariable
from .variable import Variable

if TYPE_CHECKING:
    from .base_expressions import Expression


class ExpressionCollector:
    """Walks the tree and collects handler return values."""

    def __init__(self):
        self._registry = {}

    def register(self, expr_type: type[Expression]):
        """
        Register a handler function for a specific expression type. The handler must
        return a list of results.

        :param expr_type: the type of Expression for which the handler should be used
        :return: decorator that registers the handler
        """

        def decorator(func):
            self._registry[expr_type] = func
            return func

        return decorator

    def walk(self, expr: Expression, context: Any = None) -> list[Any]:
        """
        Traverse the expression tree and apply handlers to matching types.

        :param expr: the root expression to walk
        :param context: optional context object passed to handlers
        :return: list of collected results from the handlers
        """
        return self._visit(expr, context)

    def _visit(self, expr: Expression, context: dict[str, Any]) -> list[Any]:
        """
        Recursively visit expressions and collect handler results.

        Each handler function must return a list. If multiple handlers are triggered
        during the traversal, their outputs are flattened into a single list.

        :param expr: current expression node
        :param context: context passed down to handler functions
        :return: list of results from handler invocations
        """
        if context is None:
            context = {}
        if 'ancestors' not in context:
            context['ancestors'] = []
        context['ancestors'].append(expr)

        results = []
        handler = self._registry.get(type(expr))
        if handler:
            result = handler(expr, context)
            if not isinstance(result, list):
                raise TypeError(
                    f"Handler for {type(expr).__name__} must return a list, got {type(result).__name__}"
                )
            results.extend(result)
        for child in expr.get_children():
            child_result = self._visit(child, context)
            results.extend(child_result)
        context['ancestors'].pop()
        return results


def collect_init_values(expression: Expression) -> dict[str, float]:
    collector = ExpressionCollector()

    @collector.register(Beta)
    def collect_beta(expr: Beta, context):
        if expr.is_free:
            return [(expr.name, expr.init_value)]
        return []

    collected = collector.walk(expression)
    return dict(collected)


def list_of_variables_in_expression(
    the_expression: Expression,
) -> list[Variable]:
    # Create walker
    walker = ExpressionCollector()

    @walker.register(Variable)
    def retrieve_controllers(
        expr: Variable, context: Any | None = None
    ) -> list[Variable]:
        return [expr]

    # Now use it
    return walker.walk(the_expression)


def list_of_all_betas_in_expression(
    the_expression: Expression,
) -> list[Beta]:
    # Create walker
    walker = ExpressionCollector()

    @walker.register(Beta)
    def retrieve_controllers(expr: Beta, context: Any | None = None) -> list[Beta]:
        return [expr]

    # Now use it
    return walker.walk(the_expression)


def list_of_free_betas_in_expression(
    the_expression: Expression,
) -> list[Beta]:
    # Create walker
    walker = ExpressionCollector()

    @walker.register(Beta)
    def retrieve_controllers(expr: Beta, context: Any | None = None) -> list[Beta]:
        return [expr] if expr.is_free else []

    # Now use it
    return walker.walk(the_expression)


def list_of_fixed_betas_in_expression(
    the_expression: Expression,
) -> list[Beta]:
    # Create walker
    walker = ExpressionCollector()

    @walker.register(Beta)
    def retrieve_controllers(expr: Beta, context: Any | None = None) -> list[Beta]:
        return [] if expr.is_free else [expr]

    # Now use it
    return walker.walk(the_expression)


def list_of_random_variables_in_expression(
    the_expression: Expression,
) -> list[RandomVariable]:
    # Create walker
    walker = ExpressionCollector()

    @walker.register(RandomVariable)
    def retrieve_controllers(
        expr: RandomVariable, context: Any | None = None
    ) -> list[RandomVariable]:
        return [expr]

    # Now use it
    return walker.walk(the_expression)


def list_of_draws_in_expression(
    the_expression: Expression,
) -> list[Draws]:
    # Create walker
    walker = ExpressionCollector()

    @walker.register(Draws)
    def retrieve_controllers(expr: Draws, context: Any | None = None) -> list[Draws]:
        return [expr]

    # Now use it
    return walker.walk(the_expression)



import os
import tempfile
import unittest

import pandas as pd

from biogeme.database import Database
from biogeme.expressions import Variable
from biogeme.partition import Partition
from biogeme.sampling_of_alternatives import (
    ChoiceSetsGeneration,
    CrossVariableTuple,
    SamplingContext,
)


class TestChoiceSetsGeneration(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        test_dir = tempfile.mkdtemp()
        self.test_file = os.path.join(test_dir, 'test_file.csv')

        self.individuals = pd.DataFrame({'choice': [1, 2, 1], 'age': [25, 30, 35]})
        alternatives = pd.DataFrame({'alt_id': [1, 2, 3, 4], 'cost': [10, 20, 30, 40]})
        the_partition = Partition([{1, 2}, {3, 4}])
        sample_sizes = [1, 1]
        mev_partition = Partition([{1, 3}, {2, 4}])
        mev_sample_sizes = [1, 1]
        context = SamplingContext(
            the_partition=the_partition,
            sample_sizes=sample_sizes,
            mev_partition=mev_partition,
            mev_sample_sizes=mev_sample_sizes,
            individuals=self.individuals,
            choice_column='choice',
            alternatives=alternatives,
            biogeme_file_name=self.test_file,
            id_column='alt_id',
            utility_function=Variable('choice'),
            combined_variables=[],
        )
        self.choice_set_generator = ChoiceSetsGeneration(context)

    def test_initialization(self):
        self.assertIsNotNone(self.choice_set_generator)

    def test_get_attributes_from_expression(self):
        expr = Variable('cost') + Variable('unknown_var')
        result = self.choice_set_generator.get_attributes_from_expression(expr)
        self.assertEqual(result, {'cost'})

    def test_process_row(self):
        row = self.individuals.iloc[0]
        processed_data = self.choice_set_generator.process_row(row)
        self.assertIn('choice', processed_data)
        self.assertIn('age', processed_data)

    def test_sample_and_merge(self):
        the_database = self.choice_set_generator.sample_and_merge(recycle=False)
        self.assertTrue(os.path.exists(self.test_file))

        df = pd.read_csv(self.test_file)
        expected_columns = [
            'choice',
            'age',
            'alt_id_0',
            'cost_0',
            '_log_proba_0',
            'alt_id_1',
            'cost_1',
            '_log_proba_1',
            '_MEV_alt_id_0',
            '_MEV_cost_0',
            '_MEV__mev_weight_0',
            '_MEV_alt_id_1',
            '_MEV_cost_1',
            '_MEV__mev_weight_1',
        ]
        self.assertListEqual(sorted(expected_columns), sorted(df.columns))
        os.remove(self.test_file)  # cleanup

    def test_define_new_variables(self):
        # Create a dummy database
        biogeme_data = pd.DataFrame({'var1': [1, 2, 3], 'var2': [4, 5, 6]})
        biogeme_database = Database('test_data', biogeme_data)

        # Set a sample size and dummy combined variables in the generator
        self.choice_set_generator.sample_size = 2
        self.choice_set_generator.combined_variables = [
            CrossVariableTuple(
                name='new_var', formula=Variable('var1') + Variable('var2')
            )
        ]

        self.choice_set_generator.define_new_variables(biogeme_database)

        # The new variable names should be new_var_0 and new_var_1 based on the sample size of 2
        defined_variables = biogeme_database.dataframe.columns
        self.assertIn('new_var_0', defined_variables)
        self.assertIn('new_var_1', defined_variables)


if __name__ == '__main__':
    unittest.main()



"""Convert expressions to float and vice versa

Michel Bierlaire
Tue Mar 25 18:43:36 2025
"""

from __future__ import annotations

from biogeme.exceptions import BiogemeError
from .base_expressions import ExpressionOrNumeric, Expression
from .numeric_expressions import Numeric
from .numeric_tools import is_numeric


def validate_and_convert(expression: ExpressionOrNumeric) -> Expression:
    """Validates the expression and returns the converted expression if necessary."""
    if isinstance(expression, bool):
        return Numeric(1) if expression else Numeric(0)
    if is_numeric(expression):
        return Numeric(expression)
    if not isinstance(expression, Expression):
        raise TypeError(
            f'This is not a valid expression: {expression}. It is of type {type(expression)}'
        )
    return expression


def expression_to_value(
    expression: ExpressionOrNumeric, betas: dict[str, float] | None = None
) -> float:
    """
    Convert an expression to a float value, if possible.

    :param expression: The expression to convert. Can be a boolean, a numeric value,
                       or an instance of Expression.
    :param betas: Optional dictionary of beta values used to initialize the expression,
                  if applicable.
    :return: The numerical value of the expression as a float.
    :raises TypeError: If the input is not a valid expression type.
    :raises BiogemeError: If the expression cannot be evaluated to a numeric value.
    """
    if isinstance(expression, bool) or is_numeric(expression):
        return float(expression)
    if not isinstance(expression, Expression):
        raise TypeError(f'This is not a valid expression: {expression}')
    if betas is not None:
        expression.change_init_values(betas=betas)
    try:
        value = expression.get_value()
    except BiogemeError as e:
        error_msg = f'Expression {expression} too complex to be associated with a numeric value: {e}'
        raise BiogemeError(error_msg) from e
    return value


def get_dict_values(
    the_dict: dict[int, ExpressionOrNumeric], betas: dict[str, float] | None = None
) -> dict[int, float]:
    """If the dictionary contains Expressions, they are transformed into a
    numerical expression."""

    return {
        key: expression_to_value(expression=expr, betas=betas)
        for key, expr in the_dict.items()
    }


def get_dict_expressions(
    the_dict: dict[int, ExpressionOrNumeric],
) -> dict[int, Expression]:
    """If the dictionary contains float, they are transformed into a
    numerical expression."""

    return {key: validate_and_convert(value) for key, value in the_dict.items()}



"""Implements the Multivariate Extreme Value models.

:author: Michel Bierlaire
:date: Wed Oct 25 11:35:34 2023
"""

import logging
from biogeme.expressions import (
    Expression,
    exp,
    ExpressionOrNumeric,
    LogLogit,
)
from biogeme.deprecated import deprecated

logger = logging.getLogger(__name__)


def logmev(
    util: dict[int, ExpressionOrNumeric],
    log_gi: dict[int, ExpressionOrNumeric],
    av: dict[int, ExpressionOrNumeric],
    choice: ExpressionOrNumeric,
) -> Expression:
    """Log of the choice probability for a MEV model.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param log_gi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
            (e^{V_1},\\ldots,e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param choice: id of the alternative for which the probability must be
              calculated.

    :return: log of the choice probability of the MEV model, given by

    .. math:: V_i + \\ln G_i(e^{V_1},\\ldots,e^{V_J}) -
              \\ln\\left(\\sum_j e^{V_j + \\ln G_j(e^{V_1},
              \\ldots,e^{V_J})}\\right)

    """
    h = {i: v + log_gi[i] for i, v in util.items()}
    log_p = LogLogit(h, av, choice=choice)
    return log_p


def mev(
    util: dict[int, ExpressionOrNumeric],
    log_gi: dict[int, Expression],
    av: dict[int, ExpressionOrNumeric] | None,
    choice: ExpressionOrNumeric,
):
    """Choice probability for a MEV model.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type util: dict(int:biogeme.expressions.expr.Expression)


    :param log_gi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
              (e^{V_1}, \\ldots, e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :type log_gi: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :return: Choice probability of the   MEV model, given by

    .. math:: \\frac{e^{V_i + \\ln G_i(e^{V_1},
              \\ldots,e^{V_J})}}{\\sum_j e^{V_j +
              \\ln G_j(e^{V_1},\\ldots,e^{V_J})}}

    :rtype: biogeme.expressions.expr.Expression
    """
    return exp(logmev(util, log_gi, av, choice))


def logmev_endogenous_sampling(
    util: dict[int, ExpressionOrNumeric],
    log_gi: dict[int, Expression],
    av: dict[int, ExpressionOrNumeric] | None,
    correction: dict[int, ExpressionOrNumeric],
    choice: ExpressionOrNumeric,
):
    """Log of choice probability for a MEV model, including the
    correction for endogenous sampling as proposed by `Bierlaire, Bolduc
    and McFadden (2008)`_.

    .. _`Bierlaire, Bolduc and McFadden (2008)`:
       http://dx.doi.org/10.1016/j.trb.2007.09.003

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type util: dict(int:biogeme.expressions.expr.Expression)

    :param log_gi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
                  (e^{V_1}, \\ldots, e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :type log_gi: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)


    :param correction: a dict of expressions for the correstion terms
                       of each alternative.
    :type correction: dict(int:biogeme.expressions.expr.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :return: log of the choice probability of the MEV model, given by

    .. math:: V_i + \\ln G_i(e^{V_1}, \\ldots,e^{V_J}) + \\omega_i -
              \\ln\\left(\\sum_j e^{V_j +
              \\ln G_j(e^{V_1}, \\ldots, e^{V_J})+ \\omega_j}\\right)

    where :math:`\\omega_i` is the correction term for alternative :math:`i`.

    :rtype: biogeme.expressions.expr.Expression

    """
    h = {i: v + log_gi[i] + correction[i] for i, v in util.items()}
    log_p = LogLogit(h, av, choice)
    return log_p


@deprecated(new_func=logmev_endogenous_sampling)
def logmev_endogenousSampling(
    util: dict[int, ExpressionOrNumeric],
    log_gi: dict[int, Expression],
    av: dict[int, ExpressionOrNumeric] | None,
    correction: dict[int, ExpressionOrNumeric],
    choice: ExpressionOrNumeric,
):
    pass


def mev_endogenous_sampling(
    util: dict[int, ExpressionOrNumeric],
    log_gi: dict[int, Expression],
    av: dict[int, ExpressionOrNumeric] | None,
    correction: dict[int, ExpressionOrNumeric],
    choice: ExpressionOrNumeric,
):
    """Choice probability for a MEV model, including the correction
    for endogenous sampling as proposed by
    `Bierlaire, Bolduc and McFadden (2008)`_.

    .. _`Bierlaire, Bolduc and McFadden (2008)`:
           http://dx.doi.org/10.1016/j.trb.2007.09.003

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type util: dict(int:biogeme.expressions.expr.Expression)

    :param log_gi: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}
              (e^{V_1}, \\ldots, e^{V_J})

        where :math:`G` is the MEV generating function. If an alternative
        :math:`i` is not available, then :math:`G_i = 0`.

    :type log_gi: dict(int:biogeme.expressions.expr.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.expr.Expression)


    :param correction: a dict of expressions for the correction terms
                       of each alternative.
    :type correction: dict(int:biogeme.expressions.expr.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.expr.Expression

    :return: log of the choice probability of the MEV model, given by

    .. math:: V_i + \\ln G_i(e^{V_1}, \\ldots, e^{V_J}) + \\omega_i -
              \\ln\\left(\\sum_j e^{V_j + \\ln G_j(e^{V_1},\\ldots,e^{V_J})+
              \\omega_j}\\right)

    where :math:`\\omega_i` is the correction term for alternative :math:`i`.

    :rtype: biogeme.expressions.expr.Expression

    """
    return exp(logmev_endogenous_sampling(util, log_gi, av, correction, choice))


@deprecated(new_func=mev_endogenous_sampling)
def mev_endogenousSampling(
    util: dict[int, ExpressionOrNumeric],
    log_gi: dict[int, Expression],
    av: dict[int, ExpressionOrNumeric] | None,
    correction: dict[int, ExpressionOrNumeric],
    choice: ExpressionOrNumeric,
):
    pass



"""

Calculation of individual level parameters
==========================================

Calculation of the individual level parameters for the model defined
in :ref:`plot_b05normal_mixture`.

:author: Michel Bierlaire, EPFL
:date: Mon Apr 10 12:17:12 2023

"""

import os
import pickle
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, bioDraws, MonteCarlo

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

# %%
# Parameters. The initial value is irrelevant.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation.
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# %%
# Define values for these parameters
beta_values = {
    'ASC_CAR': 0.137,
    'ASC_TRAIN': -0.402,
    'B_COST': -1.28,
    'B_TIME': -2.26,
    'B_TIME_S': 1.65,
}

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Conditional on B_TIME_RND, we have a logit model (called the kernel).
prob_chosen = models.logit(V, av, CHOICE)

# %%
# Numerator and denominator of the formula for individual parameters.
numerator = MonteCarlo(B_TIME_RND * prob_chosen)
denominator = MonteCarlo(prob_chosen)

# %%
simulate = {
    'Numerator': numerator,
    'Denominator': denominator,
    'Choice': CHOICE,
}

# %%
# The results are saved in a picke file. The next time the script is
# run, if the file exists, the results are simply loaded instead of
# being re-calcuated.
PICKLE_FILE = 'b19individual_level_parameters.pickle'
if os.path.isfile(PICKLE_FILE):
    with open(PICKLE_FILE, 'rb') as f:
        sim = pickle.load(f)
else:
    biosim = bio.BIOGEME(database, simulate)
    sim = biosim.simulate(beta_values)
    sim['Individual-level parameters'] = sim['Numerator'] / sim['Denominator']
    with open(PICKLE_FILE, 'wb') as f:
        pickle.dump(sim, f)

sim



"""Unit tests for the function call version of the formula

Michel Bierlaire
Sat Mar 29 15:31:30 2025
"""

import unittest

import numpy as np
import pandas as pd

from biogeme.calculator import (
    CallableExpression,
    CompiledFormulaEvaluator,
    function_from_compiled_formula,
)
from biogeme.database import Database
from biogeme.expressions import Beta, Variable
from biogeme.model_elements import ModelElements
from biogeme.second_derivatives import SecondDerivativesMode


class TestFunctionFromExpression(unittest.TestCase):
    def setUp(self):
        self.data = pd.DataFrame.from_dict({'x': [1.0, 2.0, 3.0]})
        self.database = Database('test', self.data)
        self.x = Variable('x')
        self.beta = Beta('beta', 1.0, None, None, 0)
        self.expression = self.beta * self.x
        self.initial_betas = {'beta': 1.0}
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        compiled_function = CompiledFormulaEvaluator(
            model_elements=model_elements,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        self.func: CallableExpression = function_from_compiled_formula(
            the_compiled_function=compiled_function, the_betas=self.initial_betas
        )

    def test_function_output_only(self):
        result = self.func(np.array([2.0]), gradient=False, hessian=False, bhhh=False)
        expected_value = 2.0 * np.sum(self.data['x'])
        self.assertAlmostEqual(result.function, expected_value, places=6)
        self.assertIsNone(result.gradient)
        self.assertIsNone(result.hessian)
        self.assertIsNone(result.bhhh)

    def test_function_with_gradient(self):
        result = self.func(np.array([3.0]), gradient=True, hessian=False, bhhh=False)
        expected_gradient = np.sum(self.data['x'])
        self.assertAlmostEqual(result.gradient[0], expected_gradient, places=6)

    def test_function_with_hessian(self):
        result = self.func(np.array([4.0]), gradient=True, hessian=True, bhhh=False)
        self.assertTrue(np.allclose(result.hessian, np.zeros((1, 1))))

    def test_function_with_bhhh(self):
        result = self.func(np.array([5.0]), gradient=True, hessian=False, bhhh=True)
        self.assertEqual(result.bhhh.shape, (1, 1))


if __name__ == '__main__':
    unittest.main()



"""

Calculation of willingness to pay
=================================

We calculate and plot willingness to pay.
Details about this example are available in Section 4 of `Bierlaire (2018)
Calculating indicators with PandasBiogeme
<http://transp-or.epfl.ch/documents/technicalReports/Bier18a.pdf>`_

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 20:57:00 2023


"""

import sys

try:
    import matplotlib.pyplot as plt

    can_plot = True
except ModuleNotFoundError:
    can_plot = False
import biogeme.biogeme as bio
import biogeme.exceptions as excep
import biogeme.results as res

from biogeme.expressions import Derive
from optima_data import database, normalized_weight
from scenarios import scenario

# %%
# Obtain the specification for the default scenario
# The definition of the scenarios is available in :ref:`scenarios`.
V, _, _, _ = scenario()

V_PT = V[0]
V_CAR = V[1]

# %%
# Calculation of the willingness to pay using derivatives.
WTP_PT_TIME = Derive(V_PT, 'TimePT') / Derive(V_PT, 'MarginalCostPT')
WTP_CAR_TIME = Derive(V_CAR, 'TimeCar') / Derive(V_CAR, 'CostCarCHF')

# %%
# Formulas to simulate.
simulate = {
    'weight': normalized_weight,
    'WTP PT time': WTP_PT_TIME,
    'WTP CAR time': WTP_CAR_TIME,
}

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, simulate)

# %%
# Read the estimation results from the file.
try:
    results = res.bioResults(pickleFile='saved_results/b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02estimation.py in order to generate '
        'the file b02estimation.pickle.'
    )

# %%
# `simulated_values` is a Pandas dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulated_values = the_biogeme.simulate(results.getBetaValues())
simulated_values

# %%
# Note the multiplication by 60 to have the valus of time per hour and not per minute.
wtpcar = (60 * simulated_values['WTP CAR time'] * simulated_values['weight']).mean()

# %%
# Calculate confidence intervals
b = results.getBetasForSensitivityAnalysis(the_biogeme.free_beta_names())

# %%
# Returns data frame containing, for each simulated value, the left
# and right bounds of the confidence interval calculated by simulation.
left, right = the_biogeme.confidenceIntervals(b, 0.9)

# %%
# Lower bounds of the confidence intervals
left
# %%
# Upper bounds of the confidence intervals
right

# %%
# Lower and upper bounds of the willingness to pay.
wtpcar_left = (60 * left['WTP CAR time'] * left['weight']).mean()
wtpcar_right = (60 * right['WTP CAR time'] * right['weight']).mean()
print(
    f'Average WTP for car: {wtpcar:.3g} ' f'CI:[{wtpcar_left:.3g}, {wtpcar_right:.3g}]'
)

# %%
# In this specific case, there are only two distinct values in the
# population: for workers and non workers
print(
    'Unique values:      ',
    [f'{i:.3g}' for i in 60 * simulated_values['WTP CAR time'].unique()],
)


# %%
# Function calculating the willingness to pay for a group.
def wtp_for_subgroup(the_filter: 'pd.Series[np.bool_]') -> tuple[float, float, float]:
    """
    Check the value for groups of the population. Define a function that
    works for any filter to avoid repeating code.

    :param the_filter: pandas filter

    :return: willingness-to-pay for car and confidence interval
    """
    size = the_filter.sum()
    sim = simulated_values[the_filter]
    total_weight = sim['weight'].sum()
    weight = sim['weight'] * size / total_weight
    _wtpcar = (60 * sim['WTP CAR time'] * weight).mean()
    _wtpcar_left = (60 * left[the_filter]['WTP CAR time'] * weight).mean()
    _wtpcar_right = (60 * right[the_filter]['WTP CAR time'] * weight).mean()
    return _wtpcar, _wtpcar_left, _wtpcar_right


# %%
# Full time workers.
aFilter = database.data['OccupStat'] == 1
w, l, r = wtp_for_subgroup(aFilter)
print(f'WTP car for workers: {w:.3g} CI:[{l:.3g}, {r:.3g}]')

# %%
# Females.
aFilter = database.data['Gender'] == 2
w, l, r = wtp_for_subgroup(aFilter)
print(f'WTP car for females: {w:.3g} CI:[{l:.3g}, {r:.3g}]')

# %%
# Males.
aFilter = database.data['Gender'] == 1
w, l, r = wtp_for_subgroup(aFilter)
print(f'WTP car for males  : {w:.3g} CI:[{l:.3g}, {r:.3g}]')

# %%
# We plot the distribution of WTP in the population. In this case,
# there are only two values
if can_plot:
    plt.hist(
        60 * simulated_values['WTP CAR time'],
        weights=simulated_values['weight'],
    )
    plt.xlabel('WTP (CHF/hour)')
    plt.ylabel('Individuals')
    plt.show()



"""

Cross-nested logit (old syntax)
===============================

Example of a cross-nested logit model, using the original syntax for nests.
Since biogeme 3.13, a new syntax, more explicit, has been adopted.
There are two nests:

 - one with existing alternatives (car and train),
 - one with public transportation alternatives (train and Swissmetro)

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:06:44 2023

"""

import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b11cnl.py')

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %% Nest parameters.
MU_EXISTING = Beta('MU_EXISTING', 1, 1, 10, 0)
MU_PUBLIC = Beta('MU_PUBLIC', 1, 1, 10, 0)

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of nests.

ALPHA_EXISTING = Beta('ALPHA_EXISTING', 0.5, 0, 1, 0)
alpha_existing = {1: ALPHA_EXISTING, 2: 0.0, 3: 1.0}

ALPHA_PUBLIC = 1 - ALPHA_EXISTING
alpha_public = {1: ALPHA_PUBLIC, 2: 1.0, 3: 0.0}

nest_existing = MU_EXISTING, alpha_existing
nest_public = MU_PUBLIC, alpha_public
nests = nest_existing, nest_public

# %%
# The choice model is a cross-nested logit, with availability conditions.
logprob = models.logcnl_avail(V, av, nests, CHOICE)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b11cnl'

# %%
# Estimate the parameters
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



import unittest

import pandas as pd

from biogeme.catalog import (
    Catalog,
    CentralController,
    Configuration,
    Controller,
    SelectedConfigurationsIterator,
    segmentation_catalogs,
)
from biogeme.database import Database
from biogeme.expressions import (
    Beta,
    NamedExpression,
    Numeric,
    Variable,
    list_of_free_betas_in_expression,
    log,
)
from biogeme.models import loglogit


class TestSelectedExpressionsIterator(unittest.TestCase):

    def setUp(self):
        # Mock configurations

        expr_1 = Numeric(1)
        expr_2 = Numeric(2)
        expr_3 = Numeric(3)
        named_expr_1 = NamedExpression(name='expr_1', expression=expr_1)
        named_expr_2 = NamedExpression(name='expr_2', expression=expr_2)
        named_expr_3 = NamedExpression(name='expr_3', expression=expr_3)
        controller = Controller(
            controller_name='test', specification_names=['expr_1', 'expr_2', 'expr_3']
        )
        catalog = Catalog(
            catalog_name='test',
            named_expressions=[named_expr_1, named_expr_2, named_expr_3],
            controlled_by=controller,
        )
        self.central_controller = CentralController(expression=catalog)
        self.configs = self.central_controller.all_configurations

    def test_iterator_initialization_with_all_configurations(self):
        iterator = SelectedConfigurationsIterator(self.central_controller)
        self.assertEqual(iterator.current_configuration in self.configs, True)

    def test_iterator_initialization_with_selected_configurations(self):
        for selected in self.configs:
            iterator = SelectedConfigurationsIterator(
                self.central_controller, {selected}
            )
            current_configuration = next(iterator)
            self.assertEqual(current_configuration, selected)

    def test_iteration_order(self):
        iterator = SelectedConfigurationsIterator(self.central_controller)
        seen = []
        for config in iterator:
            seen.append(config)
        self.assertEqual(set(seen), set(self.configs))
        self.assertEqual(len(seen), len(self.configs))

    def test_iteration_stops(self):
        iterator = SelectedConfigurationsIterator(self.central_controller)
        for _ in range(len(self.configs)):
            next(iterator)
        with self.assertRaises(StopIteration):
            next(iterator)

    def test_first_iteration_flag(self):
        iterator = SelectedConfigurationsIterator(
            self.central_controller, set(self.configs)
        )
        self.assertTrue(iterator.first)
        _ = next(iterator)
        self.assertFalse(iterator.first)

    def test_number_counter(self):
        iterator = SelectedConfigurationsIterator(
            self.central_controller, set(self.configs)
        )
        for i in range(len(self.configs)):
            next(iterator)
            self.assertEqual(iterator.number, i + 1)

    def test_value(self):
        iterator = SelectedConfigurationsIterator(
            self.central_controller, set(self.configs)
        )
        expected_values = {'test:expr_2': 2.0, 'test:expr_3': 3.0, 'test:expr_1': 1.0}

        for config in iterator:
            config_id = config.get_string_id()
            the_configuration = Configuration.from_string(config_id)
            self.central_controller.set_configuration(the_configuration)
            value = self.central_controller.expression.get_value()
            self.assertEqual(value, expected_values[config_id])


class TestIterator(unittest.TestCase):
    def setUp(self):
        data = {
            'choice': [0, 1, 2, 1, 0],
            'male': [0, 1, 0, 1, 0],
            'train_tt': [15.593820, 11.937003, 8.916655, 1.999498, 9.184978],
            'train_cost': [6.674172, 2.857336, 13.017769, 1.128232, 14.439975],
            'sm_tt': [18.771054, 0.015575, 19.844231, 12.349630, 12.233063],
            'sm_cost': [0.141326, 0.461249, 10.495493, 7.997219, 0.933313],
            'car_tt': [19.475110, 4.655427, 1.812129, 12.367720, 7.649240],
            'car_cost': [19.664618, 9.335258, 17.198808, 13.606151, 9.009985],
        }

        df = pd.DataFrame(data)

        self.database = Database(name='test', dataframe=df)
        asc_car = Beta('asc_car', 0, None, None, 0)
        asc_train = Beta('asc_train', 0, None, None, 0)
        b_time = Beta('b_time', 0, None, None, 0)
        b_cost = Beta('b_cost', 0, None, None, 0)

        # %%
        segmentation_gender = self.database.generate_segmentation(
            variable=Variable('male'), mapping={0: 'female', 1: 'male'}
        )

        # %%
        # We define catalogs with two different specifications for the
        # ASC_CAR: non segmented, and segmented.
        asc_train_catalog, asc_car_catalog = segmentation_catalogs(
            generic_name='asc',
            beta_parameters=[asc_train, asc_car],
            potential_segmentations=(segmentation_gender,),
            maximum_number=1,
        )

        # %%
        # We now define a catalog  with the log travel time as well as the travel time.

        # %%
        # First for train
        train_tt_catalog = Catalog.from_dict(
            catalog_name='train_tt_catalog',
            dict_of_expressions={
                'linear': Variable('train_tt'),
                'log': log(Variable('train_tt')),
            },
        )
        # %%
        # Then for SM. But we require that the specification is the same as
        # train by defining the same controller.
        sm_tt_catalog = Catalog.from_dict(
            catalog_name='sm_tt_catalog',
            dict_of_expressions={'linear': Variable('sm_tt'), 'log': Variable('sm_tt')},
            controlled_by=train_tt_catalog.controlled_by,
        )

        # %%
        # Definition of the utility functions with linear cost.
        v_train = (
            asc_train_catalog
            + b_time * train_tt_catalog
            + b_cost * Variable('train_cost')
        )
        v_swissmetro = b_time * sm_tt_catalog + b_cost * Variable('sm_cost')
        v_car = (
            asc_car_catalog
            + b_time * Variable('car_tt')
            + b_cost * Variable('car_cost')
        )

        # %%
        # Associate utility functions with the numbering of alternatives.
        v = {1: v_train, 2: v_swissmetro, 3: v_car}

        # %%
        # Definition of the model. This is the contribution of each
        # observation to the log likelihood function.
        self.log_probability = loglogit(v, None, Variable('choice'))

        central_controller = CentralController(
            expression=self.log_probability, maximum_number_of_configurations=10
        )
        self.the_iterator = central_controller.expression_iterator()

    def test_iterator(self):
        for named_expression in self.the_iterator:
            the_betas = list_of_free_betas_in_expression(
                the_expression=named_expression.expression
            )


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: logit

:author: Michel Bierlaire
:date: Sat Sep  9 15:28:39 2023
"""

from __future__ import annotations

import logging
from itertools import chain
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
import numpy as np
from biogeme.floating_point import JAX_FLOAT
from jax.scipy.special import logsumexp

from .base_expressions import Expression, LogitTuple
from .convert import validate_and_convert
from .jax_utils import JaxFunctionType
from ..deprecated import deprecated
from ..exceptions import BiogemeError

if TYPE_CHECKING:
    from . import ExpressionOrNumeric
logger: logging.Logger = logging.getLogger(__name__)


def index_of(key: float, keys: list[int]):
    """Function returning the index of a kex for biogeme_jax"""
    return jnp.argmax(keys == key)


class LogLogit(Expression):
    """Expression capturing the logit formula.

    It contains one formula for the target alternative, a dict of
    formula for the availabilities and a dict of formulas for the
    utilities

    """

    def __init__(
        self,
        util: dict[int, ExpressionOrNumeric],
        av: dict[int, ExpressionOrNumeric] | None,
        choice: ExpressionOrNumeric,
    ):
        """Constructor

        :param util: dictionary where the keys are the identifiers of
                     the alternatives, and the elements are objects
                     defining the utility functions.

        :type util: dict(int:biogeme.expressions.Expression)

        :param av: dictionary where the keys are the identifiers of
                   the alternatives, and the elements are object of
                   type biogeme.expressions.Expression defining the
                   availability conditions. If av is None, all the
                   alternatives are assumed to be always available

        :type av: dict(int:biogeme.expressions.Expression)

        :param choice: formula to obtain the alternative for which the
                       logit probability must be calculated.
        :type choice: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        Expression.__init__(self)
        self.util: dict[int, Expression] = {
            alt_id: validate_and_convert(util_expression)
            for (alt_id, util_expression) in util.items()
        }

        #: dict of availability formulas
        self.av: dict[int, Expression] | None = None
        if av is not None:
            self.av = {
                alt_id: validate_and_convert(avail_expression)
                for (alt_id, avail_expression) in av.items()
            }
            for i, e in self.av.items():
                self.children.append(e)
            self.av_keys = jnp.array(list(self.av.keys()), dtype=JAX_FLOAT)
            self.av_values = tuple(self.av[k] for k in self.av.keys())

        self.choice: Expression = validate_and_convert(choice)
        """expression for the chosen alternative"""

        self.children.append(self.choice)
        for i, e in self.util.items():
            self.children.append(e)

        # Convert the dict into list for biogeme_jax
        self.util_keys = jnp.array(list(self.util.keys()), dtype=JAX_FLOAT)
        self.util_values = tuple(self.util[k] for k in self.util.keys())

    def deep_flat_copy(self) -> LogLogit:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_util = {key: util.deep_flat_copy() for key, util in self.util.items()}
        copy_av = (
            {
                key: av.deep_flat_copy() if av is not None else None
                for key, av in self.av.items()
            }
            if self.av is not None
            else None
        )
        copy_choice = self.choice.deep_flat_copy()
        return type(self)(util=copy_util, av=copy_av, choice=copy_choice)

    def logit_choice_avail(self) -> list[LogitTuple]:
        result: list[LogitTuple] = list(
            chain.from_iterable(e.logit_choice_avail() for e in self.children)
        )
        if self.av is not None:
            this_tuple: LogitTuple = LogitTuple(
                choice=self.choice, availabilities=self.av
            )
            result.append(this_tuple)
        return result

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float

        :raise BiogemeError: if the chosen alternative does not correspond
            to any of the utility functions

        :raise BiogemeError: if the chosen alternative does not correspond
            to any of entry in the availability condition

        """
        choice = int(self.choice.get_value())
        if choice not in self.util:
            error_msg = (
                f'Alternative {choice} does not appear in the list '
                f'of utility functions: {self.util.keys()}'
            )
            raise BiogemeError(error_msg)
        if choice not in self.av:
            error_msg = (
                f'Alternative {choice} does not appear in the list '
                f'of availabilities: {self.av.keys()}'
            )
            raise BiogemeError(error_msg)
        if self.av[choice].get_value() == 0.0:
            return -np.log(0)
        v_chosen = self.util[choice].get_value()
        denom = 0.0
        for i, V in self.util.items():
            if self.av[i].get_value() != 0.0:
                denom += np.exp(V.get_value() - v_chosen)
        return -np.log(denom)

    @deprecated(get_value)
    def getValue(self) -> float:
        pass

    def __str__(self) -> str:
        s = f'{self.get_class_name()}[choice={self.choice}]'
        util_str = ', '.join(f'{int(i)}:{e}' for i, e in self.util.items())
        s += f'U=({util_str})'
        if self.av is None:
            s += '[always available]'
        else:
            av_str = ', '.join(f'{int(i)}:{e}' for i, e in self.av.items())
            s += f'av=({av_str})'
        return s

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a JAX-compatible function. This function computes the logit-based
        probability calculation based on availability and utility values.

        :return: A function that takes parameters, a row of the database,
            and random draws.
        """

        def get_value(
            expression: Expression,
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            """Retrieve the JAX function of an object and evaluate it."""
            jax_fn = expression.recursive_construct_jax_function(
                numerically_safe=numerically_safe
            )
            return jax_fn(parameters, one_row, the_draws, the_random_variables)

        if self.av is None:

            def the_jax_function(
                parameters: jnp.ndarray,
                one_row: jnp.ndarray,
                the_draws: jnp.ndarray,
                the_random_variables: jnp.ndarray,
            ) -> jnp.ndarray:
                """JAX-compatible function for logit probability calculation
                with availability."""

                choice_id = get_value(
                    self.choice, parameters, one_row, the_draws, the_random_variables
                )
                choice_index = index_of(choice_id, self.util_keys)

                # Compute v_chosen
                branches = tuple(
                    lambda _, V=V_expr: jnp.asarray(
                        get_value(
                            V, parameters, one_row, the_draws, the_random_variables
                        ),
                        dtype=JAX_FLOAT,
                    )
                    for V_expr in self.util_values
                )
                v_chosen = jax.lax.switch(choice_index, branches, operand=None)

                # Vectorized computation of utilities and availabilities
                all_utils = jnp.array(
                    [
                        get_value(
                            V, parameters, one_row, the_draws, the_random_variables
                        )
                        - v_chosen
                        for V in self.util_values
                    ]
                )

                # Compute the log-sum-exp safely
                result = -logsumexp(all_utils)
                return result

            return the_jax_function

        else:

            def the_jax_function(
                parameters: jnp.ndarray,
                one_row: jnp.ndarray,
                the_draws: jnp.ndarray,
                the_random_variables: jnp.ndarray,
            ) -> jnp.ndarray:
                """JAX-compatible function for logit probability calculation."""

                choice_id = get_value(
                    self.choice, parameters, one_row, the_draws, the_random_variables
                )
                choice_index = index_of(choice_id, self.util_keys)

                # Get availability of chosen alternative
                av_branches = tuple(
                    lambda _, av=av_expr: get_value(
                        av, parameters, one_row, the_draws, the_random_variables
                    )
                    for av_expr in self.av_values
                )
                chosen_avail = jax.lax.switch(choice_index, av_branches, operand=None)

                def unavailable_branch(_):
                    # If the chosen alternative is unavailable
                    return -jnp.finfo(JAX_FLOAT).max

                def available_branch(_):
                    # Compute v_chosen
                    branches = tuple(
                        lambda _, V=V_expr: jnp.asarray(
                            get_value(
                                V, parameters, one_row, the_draws, the_random_variables
                            ),
                            dtype=JAX_FLOAT,
                        )
                        for V_expr in self.util_values
                    )
                    v_chosen = jax.lax.switch(choice_index, branches, operand=None)

                    # Vectorized computation of utilities and availabilities
                    all_utils = jnp.array(
                        [
                            get_value(
                                V, parameters, one_row, the_draws, the_random_variables
                            )
                            - v_chosen
                            for V in self.util_values
                        ]
                    )
                    all_avail = jnp.array(
                        [
                            get_value(
                                A, parameters, one_row, the_draws, the_random_variables
                            )
                            for A in self.av_values
                        ]
                    )

                    masked_utils = jnp.where(all_avail != 0.0, all_utils, -jnp.inf)
                    return -logsumexp(masked_utils)

                # Conditionally compute result
                result = jax.lax.cond(
                    chosen_avail == 0.0,
                    unavailable_branch,
                    available_branch,
                    operand=None,
                )
                return result

            return the_jax_function



"""
.. _swissmetro_data:

Data preparation for Swissmetro
===============================

Data preparation and definition of the variables.

:author: Michel Bierlaire, EPFL
:date: Mon Mar  6 15:17:03 2023

"""

import pandas as pd
import biogeme.database as db
from biogeme.expressions import Variable

import os


def read_data() -> db.Database:
    """Read the data from file"""
    # Get the directory of the current file
    module_dir = os.path.dirname(__file__)

    # Construct the path to the data file
    data_file_path = os.path.join(module_dir, 'data', 'swissmetro.dat')

    # %%
    # Read the data.
    df = pd.read_csv(data_file_path, sep='\t')
    database = db.Database('swissmetro', df)
    exclude = CHOICE == 0
    # exclude = ((PURPOSE != 1) * (PURPOSE != 3) + (CHOICE == 0)) > 0
    database.remove(exclude)
    # Definition of new variables.
    _ = database.define_variable('SM_COST', SM_CO * (GA == 0))
    _ = database.define_variable('TRAIN_COST', TRAIN_CO * (GA == 0))
    _ = database.define_variable('CAR_AV_SP', CAR_AV * (SP != 0))
    _ = database.define_variable('TRAIN_AV_SP', TRAIN_AV * (SP != 0))
    _ = database.define_variable('TRAIN_TT_SCALED', TRAIN_TT / 100)
    _ = database.define_variable('TRAIN_COST_SCALED', TRAIN_COST / 100)
    _ = database.define_variable('SM_TT_SCALED', SM_TT / 100)
    _ = database.define_variable('SM_COST_SCALED', SM_COST / 100)
    _ = database.define_variable('CAR_TT_SCALED', CAR_TT / 100)
    _ = database.define_variable('CAR_CO_SCALED', CAR_CO / 100)
    return database


# Definition of the variables.
GROUP = Variable('GROUP')
SURVEY = Variable('SURVEY')
SP = Variable('SP')
ID = Variable('ID')
PURPOSE = Variable('PURPOSE')
FIRST = Variable('FIRST')
TICKET = Variable('TICKET')
WHO = Variable('WHO')
LUGGAGE = Variable('LUGGAGE')
AGE = Variable('AGE')
MALE = Variable('MALE')
INCOME = Variable('INCOME')
GA = Variable('GA')
ORIGIN = Variable('ORIGIN')
DEST = Variable('DEST')
TRAIN_AV = Variable('TRAIN_AV')
CAR_AV = Variable('CAR_AV')
SM_AV = Variable('SM_AV')
TRAIN_TT = Variable('TRAIN_TT')
TRAIN_CO = Variable('TRAIN_CO')
TRAIN_HE = Variable('TRAIN_HE')
SM_TT = Variable('SM_TT')
SM_CO = Variable('SM_CO')
SM_HE = Variable('SM_HE')
SM_SEATS = Variable('SM_SEATS')
CAR_TT = Variable('CAR_TT')
CAR_CO = Variable('CAR_CO')
CHOICE = Variable('CHOICE')
SM_COST = Variable('SM_COST')
TRAIN_COST = Variable('TRAIN_COST')
CAR_AV_SP = Variable('CAR_AV_SP')
TRAIN_AV_SP = Variable('TRAIN_AV_SP')
TRAIN_TT_SCALED = Variable('TRAIN_TT_SCALED')
TRAIN_COST_SCALED = Variable('TRAIN_COST_SCALED')
SM_TT_SCALED = Variable('SM_TT_SCALED')
SM_COST_SCALED = Variable('SM_COST_SCALED')
CAR_TT_SCALED = Variable('CAR_TT_SCALED')
CAR_CO_SCALED = Variable('CAR_CO_SCALED')



import unittest

import biogeme.biogeme as bio
import biogeme.biogeme_logging as blog
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_HE,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_HE,
    TRAIN_TT,
    read_data,
)
from biogeme.expressions import Beta, Derive
from biogeme.nests import NestsForCrossNestedLogit, OneNestForCrossNestedLogit

logger = blog.get_screen_logger(level=blog.INFO)

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)

ASC_CAR = Beta('ASC_CAR', -0.6, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -0.3, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME_SM = Beta('B_TIME_SM', -1, None, None, 0)
B_TIME_TRAIN = Beta('B_TIME_TRAIN', -1.07, None, None, 0)
B_TIME_CAR = Beta('B_TIME_CAR', -0.86, None, None, 0)
B_COST = Beta('B_COST', -0.97, None, None, 0)
B_HEADWAY_SM = Beta('B_HEADWAY_SM', -0.008, None, None, 0)
B_HEADWAY_TRAIN = Beta('B_HEADWAY_TRAIN', -0.004, None, None, 0)
GA_TRAIN = Beta('GA_TRAIN', 1.14, None, None, 0)
GA_SM = Beta('GA_SM', -0.14, None, None, 0)

MU_EXISTING = Beta('MU_EXISTING', 1.77, 1, None, 0)
MU_PUBLIC = Beta('MU_PUBLIC', 1.84, 1, None, 0)
ALPHA_EXISTING = Beta('ALPHA_EXISTING', 0.645, 0, 1, 0)
ALPHA_PUBLIC = 1.0 - ALPHA_EXISTING


SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)

V1 = (
    ASC_TRAIN
    + B_TIME_TRAIN * TRAIN_TT / 100
    + B_COST * TRAIN_COST_SCALED
    + B_HEADWAY_TRAIN * TRAIN_HE
    + GA_TRAIN * GA
)
V2 = (
    ASC_SM
    + B_TIME_SM * SM_TT_SCALED
    + B_COST * SM_COST_SCALED
    + B_HEADWAY_SM * SM_HE
    + GA_SM * GA
)
V3 = ASC_CAR + B_TIME_CAR * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests:
alpha_existing = {1: ALPHA_EXISTING, 2: 0.0, 3: 1.0}

alpha_public = {1: ALPHA_PUBLIC, 2: 1.0, 3: 0.0}

nest_existing = OneNestForCrossNestedLogit(
    nest_param=MU_EXISTING, dict_of_alpha=alpha_existing, name='existing'
)
nest_public = OneNestForCrossNestedLogit(
    nest_param=MU_PUBLIC, dict_of_alpha=alpha_public, name='public'
)
nests = NestsForCrossNestedLogit(
    choice_set=[1, 2, 3], tuple_of_nests=(nest_existing, nest_public)
)

# The choice model is a cross-nested logit, with availability conditions
log_prob = models.logcnl(V, av, nests, CHOICE)
prob1 = models.cnl(V, av, nests, 1)
gen_elas1 = Derive(prob1, 'TRAIN_TT') * TRAIN_TT / prob1
simulate = {'Prob. train': prob1, 'Elas. 1': gen_elas1}


class test_11(unittest.TestCase):
    def testEstimationAndSimulation(self):
        biogeme = bio.BIOGEME(
            database,
            log_prob,
            generate_html=False,
            generate_yaml=False,
            save_iterations=False,
        )
        biogeme.model_name = 'test_11'
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -4997.865308202665, 2)
        biosim = bio.BIOGEME(database, simulate)
        simresults = biosim.simulate(results.get_beta_values())
        self.assertAlmostEqual(
            sum(simresults['Prob. train']), 902.4604168804027, delta=1
        )


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: Derive

Michel Bierlaire
Fri May 02 2025, 13:24:27
"""

from __future__ import annotations

import logging

import jax
import jax.numpy as jnp

from .base_expressions import ExpressionOrNumeric
from .collectors import list_of_all_betas_in_expression, list_of_variables_in_expression
from .jax_utils import JaxFunctionType
from .unary_expressions import UnaryOperator
from .variable import Variable

logger = logging.getLogger(__name__)


class Derive(UnaryOperator):
    """
    Derivative with respect to a variable
    """

    def __init__(self, child: ExpressionOrNumeric, name: str):
        """Constructor

        :param child: first arithmetic expression
        :type child: biogeme.expressions.Expression
        """
        super().__init__(child)
        # Name of the elementary expression by which the derivative is taken
        self.name = name
        # Check if it is a variable or a parameter

        list_of_variables = list_of_variables_in_expression(child)
        self.variable: Variable | None = next(
            (variable for variable in list_of_variables if variable.name == name), None
        )
        if self.variable is not None:
            self.children.append(self.variable)
        list_of_betas = list_of_all_betas_in_expression(child)
        self.beta: Variable | None = next(
            (beta for beta in list_of_betas if beta.name == name), None
        )
        if self.beta is not None:
            self.children.append(self.beta)

        if self.beta is None and self.variable is None:
            logger.warning(
                f'Variable {name} does not appear in expression {child}. Derivative is trivially zero.'
            )

    def deep_flat_copy(self) -> Derive:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_child = self.child.deep_flat_copy()
        return type(self)(child=copy_child, name=self.name)

    def __str__(self) -> str:
        return f'Derive({self.child}, "{self.name}")'

    def __repr__(self) -> str:
        return f'Derive({self.child}, "{self.name}")'

    def recursive_construct_jax_function_variable(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        child_jax = self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            # Compute gradient with respect to the data row (i.e., Variable values)
            grad_wrt_row = jax.grad(
                lambda p, row, d, rv: child_jax(p, row, d, rv), argnums=1
            )
            # Get derivative w.r.t. Variable 'X' (assuming it’s index i in row)
            index = self.variable.safe_variable_id
            value = grad_wrt_row(parameters, one_row, the_draws, the_random_variables)[
                index
            ]
            return value

        return the_jax_function

    def recursive_construct_jax_function_beta(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        child_jax = self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            # Compute gradient with respect to beta
            grad_wrt_beta = jax.grad(
                lambda p, row, d, rv: child_jax(p, row, d, rv), argnums=0
            )
            # Get derivative w.r.t. Variable 'X' (assuming it’s index i in row)
            index = self.beta.safe_beta_id
            value = grad_wrt_beta(parameters, one_row, the_draws, the_random_variables)[
                index
            ]
            return value

        return the_jax_function

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        if self.beta is not None:
            return self.recursive_construct_jax_function_beta(
                numerically_safe=numerically_safe
            )
        if self.variable is not None:
            return self.recursive_construct_jax_function_variable(
                numerically_safe=numerically_safe
            )
        # Return zero function if neither beta nor variable is matched
        return lambda p, row, d, rv: jnp.array(0.0)



import logging

from biogeme.calculator import CompiledFormulaEvaluator
from biogeme.default_parameters import ParameterValue
from biogeme.model_elements import ModelElements
from biogeme.optimization import OptimizationAlgorithm
from biogeme.second_derivatives import SecondDerivativesMode
from joblib import Parallel, delayed
from tqdm import tqdm
from tqdm_joblib import tqdm_joblib

from .model_estimation import AlgorithmResults, model_estimation

logger = logging.getLogger(__name__)


def bootstrap(
    number_of_bootstrap_samples,
    the_algorithm: OptimizationAlgorithm,
    modeling_elements: ModelElements,
    parameters: dict[str, ParameterValue],
    starting_values: dict[str, float],
    second_derivatives_mode: SecondDerivativesMode,
    numerically_safe: bool,
    number_of_jobs: int,
) -> list[AlgorithmResults]:
    """
    Perform bootstrap estimation to assess the variability of model parameters.

    This function generates a specified number of bootstrap samples from the
    original dataset, estimates the model on each sample using the provided
    algorithm and parameters, and returns the collection of estimation results.

    :param number_of_bootstrap_samples: Number of bootstrap replications to perform.
    :param the_algorithm: The optimization algorithm used to estimate the model.
    :param modeling_elements: The components defining the model, including the database and log-likelihood expression.
    :param parameters: Configuration parameters used during estimation.
    :param starting_values: Dictionary of initial values for the model's free parameters.
    :param second_derivatives_mode: specifies how second derivatives are calculated.
    :param numerically_safe: improves the numerical stability of the calculations.

    :return: A list of tuples containing:
        - estimated parameter values (NumPy array),
        - diagnostic information from the optimizer (dictionary),
        - convergence status (boolean).
    """
    the_database = modeling_elements.database

    def run_one_bootstrap_estimation(_):
        bootstrap_modeling_elements = ModelElements(
            expressions=modeling_elements.expressions,
            database=the_database.bootstrap_sample(),
            number_of_draws=modeling_elements.number_of_draws,
            draws_management=None,
            user_defined_draws=modeling_elements.user_defined_draws,
            expressions_registry=None,
        )
        compiled_formula = CompiledFormulaEvaluator(
            model_elements=bootstrap_modeling_elements,
            second_derivatives_mode=second_derivatives_mode,
            numerically_safe=numerically_safe,
        )
        one_result = model_estimation(
            the_algorithm=the_algorithm,
            function_evaluator=compiled_formula,
            parameters=parameters,
            some_starting_values=starting_values,
            save_iterations_filename=None,
        )
        return one_result

    PARALLEL = True

    logger.info(f'Number of jobs for bootstrapping: {number_of_jobs}')
    if PARALLEL:
        with tqdm_joblib(
            tqdm(
                desc="Bootstraps",
                total=number_of_bootstrap_samples,
            )
        ) as progress_bar:
            results = Parallel(n_jobs=number_of_jobs)(
                delayed(run_one_bootstrap_estimation)(_)
                for _ in range(number_of_bootstrap_samples)
            )
        return results
    else:
        results = []
        for _ in tqdm(range(number_of_bootstrap_samples)):
            results.append(run_one_bootstrap_estimation(_))
        return results



""" Implementation of the pdf and CDF of common distributions

:author: Michel Bierlaire

:date: Thu Apr 23 12:01:49 2015

"""

from typing import Union
from biogeme.expressions import (
    Expression,
    log,
    exp,
    Numeric,
    MultipleSum,
    validate_and_convert,
)
from biogeme.exceptions import BiogemeError


def normalpdf(
    x: Union[float, Expression],
    mu: Union[float, Expression] = Numeric(0.0),
    s: Union[float, Expression] = Numeric(1.0),
) -> Expression:
    """
    Normal pdf

    Probability density function of a normal distribution

    .. math:: f(x;\\mu, \\sigma) =
        \\frac{1}{\\sigma \\sqrt{2\\pi}} \\exp{-\\frac{(x-\\mu)^2}{2\\sigma^2}}

    :param x: value at which the pdf is evaluated.

    :param mu: location parameter :math:`\\mu` of the Normal distribution.

    :param s: scale parameter :math:`\\sigma` of the Normal distribution.

    :note: It is assumed that :math:`\\sigma > 0`.

    :return: value of the Normal pdf.

    :raise ValueError: if :math:`\\sigma \\leq 0`.
    """
    x_expr = validate_and_convert(x)
    mu_expr = validate_and_convert(mu)
    s_expr = validate_and_convert(s)
    try:
        s_value = s_expr.get_value()
    except NotImplementedError:
        s_value = None

    if (s_value is not None) and (s_value <= 0):
        raise ValueError(f'Scale parameter must be positive and not {s_value}')

    d = -(x_expr - mu_expr) * (x_expr - mu_expr)
    n = Numeric(2.0) * s_expr * s_expr
    a = d / n
    num = exp(a)
    den = s_expr * Numeric(2.506628275)
    p = num / den
    return p


def lognormalpdf(
    x: Union[float, Expression],
    mu: Union[float, Expression] = Numeric(0.0),
    s: Union[float, Expression] = Numeric(1.0),
) -> Expression:
    """
    Log normal pdf

    Probability density function of a log normal distribution

    .. math:: f(x;\\mu, \\sigma) =
              \\frac{1}{x\\sigma \\sqrt{2\\pi}}
              \\exp{-\\frac{(\\ln x-\\mu)^2}{2\\sigma^2}}


    :param x: value at which the pdf is evaluated.

    :param mu: location parameter :math:`\\mu` of the lognormal distribution.

    :param s: scale parameter :math:`\\sigma` of the lognormal distribution.

    :note: It is assumed that :math:`\\sigma > 0`, but it is not
        verified by the code.

    :return: value of the lognormal pdf.

    """
    x_expr = validate_and_convert(x)
    mu_expr = validate_and_convert(mu)
    s_expr = validate_and_convert(s)

    try:
        x_value = x_expr.get_value()
    except (NotImplementedError, BiogemeError):
        x_value = None

    if (x_value is not None) and (x_value <= 0):
        raise ValueError(f'Argument must be positive and not {x_value}')

    try:
        s_value = s_expr.get_value()
    except (NotImplementedError, BiogemeError):
        s_value = None

    if (s_value is not None) and (s_value <= 0):
        raise ValueError(f'Scale parameter must be positive and not {s_value}')

    d = -(log(x_expr) - mu_expr) * (log(x_expr) - mu_expr)
    n = Numeric(2.0) * s_expr * s_expr
    a = d / n
    num = exp(a)
    den = x_expr * s_expr * Numeric(2.506628275)
    p = (x_expr > Numeric(0)) * num / den
    return p


def uniformpdf(
    x: Union[float, Expression],
    a: Union[float, Expression] = Numeric(-1),
    b: Union[float, Expression] = Numeric(1.0),
) -> Expression:
    """
    Uniform pdf

    Probability density function of a uniform distribution.

    .. math::  f(x;a, b) = \\left\\{ \\begin{array}{ll}
              \\frac{1}{b-a} & \\text{for } x \\in [a, b] \\\\
              0 & \\text{otherwise}\\end{array} \\right.

    :param x: argument of the pdf
    :param a: lower bound :math:`a` of the distribution. Default: -1.

    :param b: upper bound :math:`b` of the distribution. Default: 1.

    :note: It is assumed that :math:`a < b`, but it is
        not verified by the code.
    :return: value of the uniform pdf.

 """
    x_expr = validate_and_convert(x)
    a_expr = validate_and_convert(a)
    b_expr = validate_and_convert(b)
    try:
        a_value = a_expr.get_value()
    except NotImplementedError:
        a_value = None
    try:
        b_value = b_expr.get_value()
    except NotImplementedError:
        b_value = None

    if a_value is not None and b_value is not None:
        if a_value > b_value:
            raise ValueError(f'Condition {a_value} <= {b_value} is not verified.')
    result = (
        (x_expr < a_expr) * Numeric(0.0)
        + (x_expr > b_expr) * Numeric(0.0)
        + (x_expr >= a_expr) * (x_expr <= b_expr) / (b_expr - a_expr)
    )
    return result


def triangularpdf(
    x: Union[float, Expression],
    a: Union[float, Expression] = Numeric(-1.0),
    b: Union[float, Expression] = Numeric(1.0),
    c: Union[float, Expression] = Numeric(0.0),
) -> Expression:
    """
    Triangular pdf

    Probability density function of a triangular distribution

    .. math:: f(x;a, b, c) = \\left\\{ \\begin{array}{ll} 0 &
             \\text{if } x < a \\\\\\frac{2(x-a)}{(b-a)(c-a)} &
             \\text{if } a \\leq x < c \\\\\\frac{2(b-x)}{(b-a)(b-c)} &
             \\text{if } c \\leq x < b \\\\0 & \\text{if } x \\geq b.
             \\end{array} \\right.

    :param x: argument of the pdf

    :param a: lower bound :math:`a` of the distribution. Default: -1.

    :param b: upper bound :math:`b` of the distribution. Default: 1.

    :param c: mode :math:`c` of the distribution. Default: 0.

    :note: It is assumed that :math:`a <  c < b`, but it is
        not verified by the code.
    :return: value of the triangular pdf.

    """
    x_expr = validate_and_convert(x)
    a_expr = validate_and_convert(a)
    b_expr = validate_and_convert(b)
    c_expr = validate_and_convert(c)
    try:
        a_value = a_expr.get_value()
    except (NotImplementedError, BiogemeError):
        a_value = None
    try:
        b_value = b_expr.get_value()
    except (NotImplementedError, BiogemeError):
        b_value = None
    try:
        c_value = c_expr.get_value()
    except (NotImplementedError, BiogemeError):
        c_value = None

    if all(var is not None for var in (a_value, b_value, c_value)):
        if c_value <= a_value or c_value >= b_value:
            error_msg = (
                f'The following condition is not verified: a [{a_value}] < '
                f'c [{c_value} < b [{b_value}]]'
            )
            raise ValueError(error_msg)

    # x < a
    r1 = (x_expr < a_expr) * Numeric(0.0)

    # a <= x < c
    r2 = (
        (x_expr >= a_expr)
        * (x_expr < c_expr)
        * Numeric(2.0)
        * ((x_expr - a_expr) / ((b_expr - a_expr) * (c_expr - a_expr)))
    )
    #  x == c
    r3 = (x_expr == c_expr) * Numeric(2.0) / (b_expr - a_expr)

    # c < x <= b
    r4 = (
        (x_expr > c_expr)
        * (x_expr <= b_expr)
        * Numeric(2.0)
        * (b_expr - x_expr)
        / ((b_expr - a_expr) * (b_expr - c_expr))
    )

    # b < x
    r5 = (x_expr > b_expr) * Numeric(0.0)
    return MultipleSum([r1, r2, r3, r4, r5])


def logisticcdf(
    x: Union[float, Expression],
    mu: Union[float, Expression] = Numeric(0.0),
    s: Union[float, Expression] = Numeric(1.0),
) -> Expression:
    """
    Logistic CDF

    Cumulative distribution function of a logistic distribution

    .. math:: f(x;\\mu, \\sigma) = \\frac{1}
        {1+\\exp\\left(-\\frac{x-\\mu}{\\sigma} \\right)}

    :param x: value at which the CDF is evaluated.

    :param mu: location parameter :math:`\\mu` of the logistic distribution.
        Default: 0.

    :param s: scale parameter :math:`\\sigma` of the logistic distribution.
        Default: 1.

    :note: It is assumed that :math:`\\sigma > 0`, but it is
        not verified by the code.

    :return: value of the logistic CDF.

    """
    x_expr = validate_and_convert(x)
    mu_expr = validate_and_convert(mu)
    s_expr = validate_and_convert(s)
    try:
        s_value = s_expr.get_value()
    except (NotImplementedError, BiogemeError):
        s_value = None
    if (s_value is not None) and (s_value <= 0):
        raise ValueError(f'Scale parameter must be positive and not {s_value}')

    result = Numeric(1.0) / (Numeric(1.0) + exp(-(x_expr - mu_expr) / s_expr))
    return result



import os
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from biogeme.database import Database
from biogeme.expressions import Beta, Variable, exp
from biogeme.results_processing import EstimationResults
from biogeme.tools import simulate


class DummyResults(EstimationResults):
    def __init__(self, beta_values):
        self.beta_values = beta_values

    def get_beta_values(self):
        return self.beta_values


class TestSimulateFunction(unittest.TestCase):

    def setUp(self):
        # Minimal dummy database
        self.data = pd.DataFrame({'x': [1, 2, 3]})
        self.db = Database('testdb', self.data)

        # Dummy expression
        self.beta = Beta('beta', 1.0, None, None, 0)
        self.expressions = {'exp_beta_x': exp(self.beta * Variable('x'))}

        # Dummy estimation result
        self.results = DummyResults(beta_values={self.beta.name: 2.0})

    def test_simulation_returns_dataframe(self):
        df = simulate(self.db, self.expressions, self.results)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(len(df), len(self.data))
        self.assertIn('exp_beta_x', df.columns)

    def test_simulation_csv_output(self):
        filename = 'test_output.csv'
        try:
            df = simulate(
                self.db, self.expressions, self.results, csv_filename=filename
            )
            self.assertTrue(os.path.exists(filename))
            df_loaded = pd.read_csv(filename)
            assert_frame_equal(df, df_loaded, check_dtype=False)
        finally:
            if os.path.exists(filename):
                os.remove(filename)

    def test_simulation_with_invalid_expression(self):
        with self.assertRaises(Exception):
            simulate(self.db, {'bad': 'not an expression'}, self.results)


if __name__ == '__main__':
    unittest.main()



from .function_call import CallableExpression, function_from_compiled_formula
from .multiple_formula import MultiRowEvaluator
from .simple_formula import (
    create_function_simple_expression,
    evaluate_simple_expression_per_row,
)
from .single_formula import (
    CompiledFormulaEvaluator,
    calculate_single_formula,
    evaluate_expression,
    evaluate_formula,
    evaluate_model_per_row,
    get_value_and_derivatives,
    get_value_c,
)



from .catalog import Catalog
from .catalog_iterator import SelectedConfigurationsIterator
from .central_controller import (
    CentralController,
    count_number_of_specifications,
    extract_multiple_expressions_controllers,
)
from .configuration import Configuration, SelectionTuple
from .controller import Controller, ControllerOperator
from .generic_alt_specific_catalog import generic_alt_specific_catalogs
from .segmentation_catalog import SegmentedParameters, segmentation_catalogs
from .specification import Specification



"""

biogeme.filenames
=================

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Wed Nov 22 13:47:01 2023
"""

# %%
import biogeme.version as ver
import biogeme.filenames as fl

# %%
# Version of Biogeme.
print(ver.getText())

# %%
# The role of this function is to obtain the name of a file that does not exist.
the_name = fl.get_new_file_name('test', 'dat')
the_name

# %%
# Now, let's create that file, and call the function again. A suffix
# with a number is appended to the name of the file, before its
# extension.
open(the_name, 'a').close()

# %%
the_name = fl.get_new_file_name('test', 'dat')
the_name

# %%
# If we do it again, the number is incremented.
open(the_name, 'a').close()

# %%
the_name = fl.get_new_file_name('test', 'dat')
the_name



"""Specifies the way second derivatives have to be calculated

Michel Bierlaire
Tue Jun 24 2025, 15:16:20
"""

from enum import Enum


class SecondDerivativesMode(Enum):
    ANALYTICAL = "analytical"
    FINITE_DIFFERENCES = "finite_differences"
    NEVER = "never"



'''
Test the configuration module

:author: Michel Bierlaire
:date: Sat Apr  8 17:57:33 2023Fri Mar  3 17:48:36 2023

'''

import unittest

from biogeme.catalog import Configuration, SelectionTuple
from biogeme.exceptions import BiogemeError


class TestConfiguration(unittest.TestCase):
    def test_ctor(self):
        s11 = SelectionTuple(controller='catalog1', selection='item1')
        s12 = SelectionTuple(controller='catalog1', selection='item2')
        s21 = SelectionTuple(controller='catalog2', selection='choice1')
        s22 = SelectionTuple(controller='catalog2', selection='choice2')
        with self.assertRaises(BiogemeError):
            _ = Configuration([s11, s12])
        the_conf = Configuration([s11, s22])
        the_id = the_conf.get_string_id()
        the_correct_id = 'catalog1:item1;catalog2:choice2'
        self.assertEqual(the_id, the_correct_id)

        the_empty_conf = Configuration()
        with self.assertRaises(BiogemeError):
            the_empty_conf.selections = [s11, s12]
        the_empty_conf.selections = [s11, s21]
        the_id = the_empty_conf.get_string_id()
        the_correct_id = 'catalog1:item1;catalog2:choice1'
        self.assertEqual(the_id, the_correct_id)

        the_correct_id = 'catalog1:item1;catalog2:choice1'
        another_conf = Configuration.from_string(the_id)
        the_id = another_conf.get_string_id()
        self.assertEqual(the_id, the_correct_id)

        the_dict = {'catalog1': 'item1', 'catalog2': 'choice2'}
        the_correct_id = 'catalog1:item1;catalog2:choice2'
        yet_another_conf = Configuration.from_dict(the_dict)
        the_id = yet_another_conf.get_string_id()
        self.assertEqual(the_id, the_correct_id)

        conf1 = Configuration([s11])
        conf2 = Configuration([s22])
        combined_conf = Configuration.from_tuple_of_configurations((conf1, conf2))
        the_id = combined_conf.get_string_id()
        the_correct_id = 'catalog1:item1;catalog2:choice2'
        self.assertEqual(the_id, the_correct_id)

    def test_set_of_catalogs(self):
        s11 = SelectionTuple(controller='catalog1', selection='item1')
        s21 = SelectionTuple(controller='catalog2', selection='choice1')
        the_conf = Configuration([s11, s21])
        the_set = the_conf.set_of_controllers()
        the_correct_set = {'catalog1', 'catalog2'}
        self.assertSetEqual(the_set, the_correct_set)

    def test_equal(self):
        s11 = SelectionTuple(controller='catalog1', selection='item1')
        s21 = SelectionTuple(controller='catalog2', selection='choice1')
        s22 = SelectionTuple(controller='catalog2', selection='choice2')
        conf1 = Configuration([s11, s21])
        conf2 = Configuration([s11, s21])
        conf3 = Configuration([s11, s22])
        self.assertEqual(conf1, conf2)
        self.assertNotEqual(conf1, conf3)

    def test_get_selection(self):
        s11 = SelectionTuple(controller='catalog1', selection='item1')
        s21 = SelectionTuple(controller='catalog2', selection='choice1')
        the_conf = Configuration([s11, s21])
        cat1 = the_conf.get_selection('catalog1')
        self.assertEqual(cat1, 'item1')
        cat2 = the_conf.get_selection('catalog2')
        self.assertEqual(cat2, 'choice1')
        cat3 = the_conf.get_selection('catalog3')
        self.assertIsNone(cat3)


if __name__ == '__main__':
    unittest.main()



"""

Simulation of a mixture model
=============================

Simulation of the mixture model, with estimation of the integration error.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:47:42 2023

"""

import sys
import numpy as np
import biogeme.biogeme as bio
from biogeme import models
import biogeme.results as res
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, bioDraws, MonteCarlo

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

try:
    import matplotlib.pyplot as plt

    PLOT = True
except ModuleNotFoundError:
    print('Install matplotlib to see the distribution of integration errors.')
    print('pip install matplotlib')
    PLOT = False

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation.
B_TIME = Beta('B_TIME', 0, None, None, 0)

# %%
# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# The estimation results are read from the pickle file.
try:
    results = res.bioResults(pickleFile='saved_results/b05_estimation_results.pickle')
except BiogemeError:
    print(
        'Run first the script 05normalMixture.py in order to generate the '
        'file 05normalMixture.pickle.'
    )
    sys.exit()

# %%
# Conditional to B_TIME_RND, we have a logit model (called the kernel)
prob = models.logit(V, av, CHOICE)

# %%
# We calculate the integration error. Note that this formula assumes
# independent draws, and is not valid for Halton or antithetic draws.
integral = MonteCarlo(prob)
integralSquare = MonteCarlo(prob * prob)
variance = integralSquare - integral * integral
error = (variance / 2.0) ** 0.5

# %%
# And the value of the individual parameters.
numerator = MonteCarlo(B_TIME_RND * prob)
denominator = integral

# %%
simulate = {
    'Numerator': numerator,
    'Denominator': denominator,
    'Integral': integral,
    'Integration error': error,
}

# %%
# Create the Biogeme object.
biosim = bio.BIOGEME(database, simulate)
biosim.modelName = 'b05normal_mixture_simul'

# %%
# Simulate the requested quantities. The output is a Pandas data frame.
simresults = biosim.simulate(results.getBetaValues())

# %%
# 95% confidence interval on the log likelihood.
simresults['left'] = np.log(
    simresults['Integral'] - 1.96 * simresults['Integration error']
)
simresults['right'] = np.log(
    simresults['Integral'] + 1.96 * simresults['Integration error']
)

# %%
print(f'Log likelihood: {np.log(simresults["Integral"]).sum()}')

# %%
print(
    f'Integration error for {biosim.number_of_draws} draws: '
    f'{simresults["Integration error"].sum()}'
)

# %%
print(f'In average {simresults["Integration error"].mean()} per observation.')

# %%
# 95% confidence interval
print(
    f'95% confidence interval: [{simresults["left"].sum()} - '
    f'{simresults["right"].sum()}]'
)

# %%
# Post processing to obtain the individual parameters.
simresults['beta'] = simresults['Numerator'] / simresults['Denominator']

# %%
# Plot the histogram of individual parameters
if PLOT:
    simresults['beta'].plot(kind='hist', density=True, bins=20)


# %%
# Plot the general distribution of beta
def normalpdf(val, mu=0.0, std=1.0):
    """
    Calculate the pdf of the normal distribution, for plotting purposes.

    """
    d = -(val - mu) * (val - mu)
    n = 2.0 * std * std
    a = d / n
    num = np.exp(a)
    den = std * 2.506628275
    p = num / den
    return p


# %%
betas = results.getBetaValues(['B_TIME', 'B_TIME_S'])
x = np.arange(simresults['beta'].min(), simresults['beta'].max(), 0.01)

# %%
if PLOT:
    plt.plot(x, normalpdf(x, betas['B_TIME'], betas['B_TIME_S']), '-')
    plt.show()



"""
Test the nests module

:author: Michel Bierlaire
:date: Thu Oct  5 16:13:40 2023
"""

import unittest
from unittest.mock import MagicMock

from biogeme.nests import (
    OneNestForNestedLogit,
    NestsForNestedLogit,
    OneNestForCrossNestedLogit,
    NestsForCrossNestedLogit,
    get_nest,
    Nests,
)
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta, Expression, Numeric, get_dict_values


class TestOneNestForNestedLogit(unittest.TestCase):
    def test_default_nest_name(self):
        nest = OneNestForNestedLogit(Beta('mu', 1, 1, None, 0), [1, 2, 3])

        # Assuming nest_name should be None by default
        self.assertIsNone(nest.name)

    def test_from_tuple(self):
        """Test creation from tuple."""
        nest_param = Numeric(0.5)
        list_of_alternatives = [1, 2, 3]
        tuple_input = (nest_param, list_of_alternatives)
        nest = OneNestForNestedLogit.from_tuple(tuple_input)
        self.assertEqual(repr(nest.nest_param), repr(nest_param))
        self.assertEqual(nest.list_of_alternatives, list_of_alternatives)

    def test_intersection(self):
        """Test intersection method."""
        nest1 = OneNestForNestedLogit(Numeric(0.5), [1, 2, 3])
        nest2 = OneNestForNestedLogit(Numeric(0.5), [2, 3, 4])
        self.assertEqual(nest1.intersection(nest2), {2, 3})


class TestNestsForNestedLogitExtended(unittest.TestCase):
    def test_nest_name_setting(self):
        choice_set = [1, 2, 3, 4, 5, 6]
        nests_tuple = (
            OneNestForNestedLogit(
                nest_param=Beta('mu_1', 1, 1, None, 0), list_of_alternatives=[1, 2]
            ),
            OneNestForNestedLogit(
                nest_param=Beta('mu_2', 1, 1, None, 0),
                list_of_alternatives=[3, 4],
                name="Nest2",
            ),
            OneNestForNestedLogit(
                nest_param=Beta('mu_3', 1, 1, None, 0), list_of_alternatives=[5, 6]
            ),
        )

        nests_for_nested_logit = NestsForNestedLogit(
            choice_set=choice_set, tuple_of_nests=nests_tuple
        )

        # Assuming that nest names are set in the NestsForNestedLogit constructor
        self.assertEqual(nests_for_nested_logit.tuple_of_nests[0].name, "nest_1")
        self.assertEqual(nests_for_nested_logit.tuple_of_nests[1].name, "Nest2")
        self.assertEqual(nests_for_nested_logit.tuple_of_nests[2].name, "nest_3")

    def test_check_union_extra_values(self):
        choice_set = [1, 2, 3, 4]
        nests = (
            OneNestForNestedLogit(Beta('mu_1', 1, 1, None, 0), [1, 2], 'Nest1'),
            OneNestForNestedLogit(Beta('mu_2', 1, 1, None, 0), [3, 4, 5, 6], 'Nest2'),
            # Notice that 5 and 6 are extra here
        )

        with self.assertRaisesRegex(BiogemeError, '5, 6'):
            nests_for_nested_logit = NestsForNestedLogit(
                choice_set=choice_set, tuple_of_nests=nests
            )

    def test_check_intersection_with_overlapping_nests(self):
        choice_set = [1, 2, 3, 4, 5, 6]
        nests = (
            OneNestForNestedLogit(Beta('mu_1', 1, 1, None, 0), [1, 2, 3], "Nest1"),
            OneNestForNestedLogit(Beta('mu_2', 1, 1, None, 0), [3, 4, 5], "Nest2"),
            # Notice that 3 is common in both nests
        )

        nests_for_nested_logit = NestsForNestedLogit(
            choice_set=choice_set, tuple_of_nests=nests
        )

        # Test that check_intersection identifies the intersection
        is_valid, msg = nests_for_nested_logit.check_intersection()
        self.assertFalse(is_valid)
        self.assertIn('3', msg)

    def test_old_initialization(self):
        mu_1 = Beta('mu_1', 1, 1, 10, 0)
        alternatives_1 = [1, 3]
        nest_1 = mu_1, alternatives_1
        the_nest_1 = OneNestForNestedLogit.from_tuple(nest_1)
        self.assertIs(the_nest_1.nest_param, mu_1)
        self.assertListEqual(the_nest_1.list_of_alternatives, alternatives_1)
        mu_2 = Beta('mu_2', 1, 1, 10, 0)
        alternatives_2 = [2]
        nest_2 = mu_2, alternatives_2
        the_nest_2 = OneNestForNestedLogit.from_tuple(nest_2)
        nests = NestsForNestedLogit(
            choice_set=[1, 2, 3], tuple_of_nests=(nest_1, nest_2)
        )
        for nest in nests:
            self.assertIsInstance(nest, OneNestForNestedLogit)


class TestNestsForCrossNestedLogit(unittest.TestCase):

    def setUp(self):
        choice_set = [1, 2, 3, 4]
        dict_of_alpha = {1: Numeric(0.2), 2: Numeric(0.3)}
        nest_param = Numeric(0.5)
        self.nests = (OneNestForCrossNestedLogit(nest_param, dict_of_alpha, "nest1"),)
        self.model = NestsForCrossNestedLogit(choice_set, self.nests)

    def test_get_alpha_dict(self):
        """Test that alpha dict correctly maps alternative IDs to alpha expressions."""
        expected_alpha_dict = {'nest1': Numeric(0.2)}
        alpha_dict = self.model.get_alpha_dict(1)
        self.assertIsInstance(alpha_dict['nest1'], Numeric)
        self.assertEqual(expected_alpha_dict.keys(), alpha_dict.keys())
        self.assertEqual(
            expected_alpha_dict['nest1'].get_value(),
            alpha_dict['nest1'].get_value(),
        )

    def test_one_nest(self):
        alpha_1 = Beta('alpha_1', 0, 0, 1, 0)
        alpha_2 = Beta('alpha_2', 0, 0, 1, 0)
        nest_1 = OneNestForCrossNestedLogit(
            nest_param=Beta('mu', 1, 1, None, 0), dict_of_alpha={1: alpha_1, 2: alpha_2}
        )
        is_fixed = nest_1.all_alpha_fixed()
        self.assertFalse(is_fixed)

        nest_2 = OneNestForCrossNestedLogit(
            nest_param=Beta('mu', 1, 1, None, 0), dict_of_alpha={1: 0.5, 2: 0.5}
        )
        is_fixed = nest_2.all_alpha_fixed()
        self.assertTrue(is_fixed)

    def test_get_alpha_dict_2(self):
        mu_param_1 = Beta('mu_param_1', 1, 1, 10, 0)
        alpha_param_1 = Beta('alpha_param_1', 0.5, 0, 1, 0)
        alpha_dict_1 = {1: alpha_param_1, 3: 1.0}
        the_nest_1 = OneNestForCrossNestedLogit(
            nest_param=mu_param_1, dict_of_alpha=alpha_dict_1, name='nest_1'
        )
        mu_param_2 = Beta('mu_param_2', 1, 1, 10, 0)
        alpha_param_2 = Beta('alpha_param_2', 0.5, 0, 1, 0)
        alpha_dict_2 = {1: alpha_param_2, 2: 1.0}
        the_nest_2 = OneNestForCrossNestedLogit(
            nest_param=mu_param_2, dict_of_alpha=alpha_dict_2, name='nest_2'
        )

        the_nests = NestsForCrossNestedLogit(
            choice_set=[1, 2, 3], tuple_of_nests=(the_nest_1, the_nest_2)
        )
        alpha_dict = the_nests.get_alpha_dict(alternative_id=1)
        expected_dict = {'nest_1': alpha_param_1, 'nest_2': alpha_param_2}
        self.assertDictEqual(alpha_dict, expected_dict)
        alpha_dict = get_dict_values(
            the_dict=the_nests.get_alpha_dict(alternative_id=2)
        )
        expected_dict = {'nest_1': 0, 'nest_2': 1}
        self.assertDictEqual(alpha_dict, expected_dict)

    def test_old_initialization(self):
        mu_param_1 = Beta('mu_param_1', 1, 1, 10, 0)
        alpha_param_1 = Beta('alpha_param_1', 0.5, 0, 1, 0)
        alpha_dict_1 = {1: alpha_param_1, 2: Numeric(0.0), 3: Numeric(1.0)}
        nest_1 = mu_param_1, alpha_dict_1
        the_nest_1 = OneNestForCrossNestedLogit.from_tuple(nest_1)
        self.assertIs(the_nest_1.nest_param, mu_param_1)
        self.assertEqual(
            set(the_nest_1.dict_of_alpha.keys()),
            set(alpha_dict_1.keys()),
            "Dictionaries have different keys",
        )
        for key in alpha_dict_1:
            self.assertEqual(
                repr(the_nest_1.dict_of_alpha[key]), repr(alpha_dict_1[key])
            )
        mu_param_2 = Beta('mu_param_2', 1, 1, 10, 0)
        alpha_param_2 = Beta('alpha_param_2', 0.5, 0, 1, 0)
        alpha_dict_2 = {1: alpha_param_2, 2: 1.0, 3: 0.0}
        nest_2 = mu_param_2, alpha_dict_2
        the_nest_2 = OneNestForCrossNestedLogit.from_tuple(nest_2)

        the_nests = NestsForCrossNestedLogit(
            choice_set=[1, 2, 3], tuple_of_nests=(nest_1, nest_2)
        )
        for nest in the_nests:
            self.assertIsInstance(nest, OneNestForCrossNestedLogit)


class TestOneNestForCrossNestedLogit(unittest.TestCase):
    def setUp(self):
        self.nest_param = Numeric(0.5)
        self.dict_of_alpha = {1: Numeric(0.2), 2: Numeric(0.3)}
        self.nest = OneNestForCrossNestedLogit(self.nest_param, self.dict_of_alpha)

    def test_from_tuple_old_syntax(self):
        """Test creating an instance from the old tuple syntax."""
        nest_param = Numeric(0.5)
        dict_of_alpha = {1: Numeric(0.3), 2: Numeric(0.7)}
        tuple_input = (nest_param, dict_of_alpha)
        nest = OneNestForCrossNestedLogit.from_tuple(tuple_input)
        self.assertEqual(repr(nest.nest_param), repr(nest_param))
        self.assertTrue(
            all(isinstance(alpha, Numeric) for alpha in nest.dict_of_alpha.values())
        )

    def test_all_alpha_fixed_with_mixed_types(self):
        """Test checking if all alphas are fixed when they are mixed types."""
        nest_param = Numeric(0.5)
        dict_of_alpha = {1: 0.3, 2: Numeric(0.7)}
        nest = OneNestForCrossNestedLogit(nest_param, dict_of_alpha)
        self.assertTrue(nest.all_alpha_fixed())

    def test_all_alpha_fixed_with_expressions(self):
        """Test checking if all alphas are fixed when they are expressions."""
        nest_param = Numeric(0.5)
        dict_of_alpha = {1: Numeric(0.3), 2: Numeric(0.7)}
        nest = OneNestForCrossNestedLogit(nest_param, dict_of_alpha)
        self.assertTrue(nest.all_alpha_fixed())
        dict_of_alpha = {1: Numeric(0.3), 2: Beta('alpha', 0, None, None, 0)}
        nest = OneNestForCrossNestedLogit(nest_param, dict_of_alpha)
        self.assertFalse(nest.all_alpha_fixed())
        dict_of_alpha = {1: Numeric(0.3), 2: Numeric(1) + Numeric(2)}
        nest = OneNestForCrossNestedLogit(nest_param, dict_of_alpha)
        self.assertFalse(nest.all_alpha_fixed())

    def test_post_init_list_of_alternatives(self):
        """Test if list of alternatives is correctly initialized after post init."""
        nest_param = 0.5  # Using float for simplicity
        dict_of_alpha = {1: 0.3, 2: 0.7, 3: 0.5}  # Using floats for simplicity
        nest = OneNestForCrossNestedLogit(nest_param, dict_of_alpha)
        self.assertListEqual(nest.list_of_alternatives, list(dict_of_alpha.keys()))

    def test_from_tuple(self):
        """Test creation from the old tuple syntax."""
        tuple_input = (self.nest_param, self.dict_of_alpha)
        nest = OneNestForCrossNestedLogit.from_tuple(tuple_input)
        self.assertEqual(repr(nest.nest_param), repr(self.nest_param))
        self.assertEqual(nest.dict_of_alpha, self.dict_of_alpha)


class TestGetNest(unittest.TestCase):
    def setUp(self):
        self.expression = Numeric(0.5)
        self.list_of_alternatives = [1, 2, 3]

    def test_get_nest_with_OneNestForNestedLogit(self):
        """Test that the function returns the input when it's already an instance of OneNestForNestedLogit."""
        nest = OneNestForNestedLogit(self.expression, self.list_of_alternatives)
        result = get_nest(nest)
        self.assertIs(result, nest)

    def test_get_nest_with_old_tuple(self):
        """Test that the function converts old tuple-based nest definitions into OneNestForNestedLogit instances."""
        old_nest = (self.expression, self.list_of_alternatives)
        result = get_nest(old_nest)
        self.assertIsInstance(result, OneNestForNestedLogit)
        self.assertEqual(repr(result.nest_param), repr(self.expression))
        self.assertEqual(result.list_of_alternatives, self.list_of_alternatives)

    def test_get_nest_with_invalid_type(self):
        """Test that the function raises a TypeError when the input is neither OneNestForNestedLogit nor a tuple."""
        invalid_nest = 'not a valid nest definition'
        with self.assertRaises(TypeError) as context:
            get_nest(invalid_nest)
        self.assertIn('does not represent a nest', str(context.exception))


class TestNests(unittest.TestCase):
    def test_constructor_correct_initialization(self):
        choice_set = [1, 2, 3, 4]
        nest1 = MagicMock(name='Nest1', list_of_alternatives=[1, 2])
        nest2 = MagicMock(name='Nest2', list_of_alternatives=[3, 4])
        nests = Nests(choice_set, (nest1, nest2))
        self.assertEqual(len(nests.mev_alternatives), 4)
        self.assertTrue(nests.check_names())

    def test_constructor_with_unnamed_nests(self):
        choice_set = [1, 2]
        nest = MagicMock(list_of_alternatives=[1, 2])
        nest.name = None
        nests = Nests(choice_set, (nest,))
        self.assertEqual(nest.name, 'nest_1')

    def test_constructor_with_invalid_alternatives(self):
        choice_set = [1, 2]
        nest = MagicMock(name='Nest', list_of_alternatives=[3])
        with self.assertRaises(BiogemeError) as cm:
            Nests(choice_set, (nest,))
        self.assertIn('not in the choice set', str(cm.exception))

    def test_getitem_valid_index(self):
        nest = MagicMock(name='Nest', list_of_alternatives=[1, 2])
        nests = Nests([1, 2], (nest,))
        self.assertEqual(nests[0], nest)

    def test_getitem_index_out_of_bounds(self):
        nest = MagicMock(name='Nest', list_of_alternatives=[1, 2])
        nests = Nests([1, 2], (nest,))
        with self.assertRaises(IndexError):
            _ = nests[1]

    def test_iteration(self):
        nest1 = MagicMock(name='Nest1', list_of_alternatives=[1])
        nest2 = MagicMock(name='Nest2', list_of_alternatives=[2])
        nests = Nests([1, 2], (nest1, nest2))
        for i, nest in enumerate(nests):
            self.assertEqual(nest, [nest1, nest2][i])

    def test_check_union(self):
        choice_set = [1, 2, 3, 4, 5]
        nest1 = MagicMock(name='Nest1', list_of_alternatives=[1, 2])
        nest2 = MagicMock(name='Nest2', list_of_alternatives=[3])
        nests = Nests(choice_set, (nest1, nest2))
        nests.alone = set()
        result, message = nests.check_union()
        self.assertFalse(result)
        self.assertIn('not in any nest', message)


if __name__ == '__main__':
    unittest.main()
